# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import re
from typing import List, Tuple

import click

from lean.components.util.logger import Logger
from lean.models.api import QCParameter
from lean.models.logger import Option
from lean.models.optimizer import (OptimizationConstraint, OptimizationConstraintOperator, OptimizationExtremum,
                                   OptimizationParameter, OptimizationTarget)
from lean.models.pydantic import WrappedBaseModel


class NodeType(WrappedBaseModel):
    name: str
    ram: int
    cores: int
    price: float
    min_nodes: int
    max_nodes: int
    default_nodes: int


# The nodes that are available in the cloud
# Copied from ViewsOptimization.NodeTypes in js/views/ViewsOptimization.js
available_nodes = [
    NodeType(name="O2-8",
             ram=8,
             cores=2,
             price=0.15,
             min_nodes=1,
             max_nodes=24,
             default_nodes=12),
    NodeType(name="O4-12",
             ram=12,
             cores=4,
             price=0.3,
             min_nodes=1,
             max_nodes=12,
             default_nodes=6),
    NodeType(name="O8-16",
             ram=16,
             cores=8,
             price=0.6,
             min_nodes=1,
             max_nodes=6,
             default_nodes=3)
]


class OptimizerConfigManager:
    """The OptimizationConfigurer contains methods to interactively configure parts of the optimizer."""

    def __init__(self, logger: Logger) -> None:
        """Creates a new OptimizationConfigurer instance.

        :param logger: the logger to use when printing messages
        """
        self._logger = logger

        # The targets that are available in the cloud
        self.available_targets = [
            ("TotalPerformance.PortfolioStatistics.SharpeRatio", "Sharpe Ratio"),
            ("TotalPerformance.PortfolioStatistics.CompoundingAnnualReturn", "Compounding Annual Return"),
            ("TotalPerformance.PortfolioStatistics.ProbabilisticSharpeRatio", "Probabilistic Sharpe Ratio"),
            ("TotalPerformance.PortfolioStatistics.Drawdown", "Drawdown")
        ]

    def configure_strategy(self, cloud: bool) -> str:
        """Asks the user for the optimization strategy to use.

        :param cloud: True if the optimization will be ran in the cloud, False if not
        :return: the class name of the optimization strategy to use
        """
        options = [
            Option(id="QuantConnect.Optimizer.Strategies.GridSearchOptimizationStrategy", label="Grid Search")
        ]

        if not cloud:
            options.append(
                Option(id="QuantConnect.Optimizer.Strategies.EulerSearchOptimizationStrategy", label="Euler Search")
            )

        return self._logger.prompt_list("Select the optimization strategy to use", options)

    def configure_target(self) -> OptimizationTarget:
        """Asks the user for the optimization target.

        :return: the chosen optimization target
        """
        # Create a list of options containing a "<target> (min)" and "<target> (max)" option for every target
        options = list(itertools.product(self.available_targets,
                                         [OptimizationExtremum.Minimum, OptimizationExtremum.Maximum]))
        options = [Option(id=OptimizationTarget(target=option[0][0], extremum=option[1]),
                          label=f"{option[0][1]} ({option[1]})") for option in options]

        return self._logger.prompt_list("Select an optimization target", options)

    def configure_parameters(self, project_parameters: List[QCParameter], cloud: bool) -> List[OptimizationParameter]:
        """Asks the user which parameters need to be optimized and with what constraints.

        :param project_parameters: the parameters of the project that will be optimized
        :param cloud: True if the optimization will be ran in the cloud, False if not
        :return: the chosen optimization parameters
        """
        results: List[OptimizationParameter] = []

        for parameter in project_parameters:
            if cloud and len(results) == 2:
                self._logger.warn(f"You can optimize up to 2 parameters in the cloud, skipping '{parameter.key}'")
                continue

            if not click.confirm(f"Should the '{parameter.key}' parameter be optimized?", default=True):
                continue

            minimum = click.prompt(f"Minimum value for '{parameter.key}'", type=click.FLOAT)
            maximum = click.prompt(f"Maximum value for '{parameter.key}'", type=click.FloatRange(min=minimum))
            step_size = click.prompt(f"Step size for '{parameter.key}'", type=click.FloatRange(min=0.0), default=1.0)

            results.append(OptimizationParameter(name=parameter.key, min=minimum, max=maximum, step=step_size))

        return results

    def configure_constraints(self) -> List[OptimizationConstraint]:
        """Asks the user for the optimization constraints.

        :return: the chosen optimization constraints
        """
        self._logger.info("Constraints can be used to filter out backtests from the results")
        self._logger.info("When a backtest doesn't comply with the constraints it is dropped from the results")
        self._logger.info("Example constraint: Drawdown < 0.25 (Drawdown less than 25%)")

        results: List[OptimizationConstraint] = []

        while True:
            results_str = ", ".join([str(result) for result in results])
            results_str = results_str or "None"
            self._logger.info(f"Current constraints: {results_str}")

            if not click.confirm("Do you want to add a constraint?", default=False):
                return results

            target_options = [Option(id=target[0], label=target[1]) for target in self.available_targets]
            target = self._logger.prompt_list("Select a constraint target", target_options)

            operator = self._logger.prompt_list("Select a constraint operator (<value> will be asked after this)", [
                Option(id=OptimizationConstraintOperator.Less, label="Less than <value>"),
                Option(id=OptimizationConstraintOperator.LessOrEqual, label="Less than or equal to <value>"),
                Option(id=OptimizationConstraintOperator.Greater, label="Greater than <value>"),
                Option(id=OptimizationConstraintOperator.GreaterOrEqual, label="Greater than or equal to <value>"),
                Option(id=OptimizationConstraintOperator.Equals, label="Equal to <value>"),
                Option(id=OptimizationConstraintOperator.NotEqual, label="Not equal to <value>")
            ])

            value = click.prompt("Set the <value> for the selected operator", type=click.FLOAT)

            results.append(OptimizationConstraint(**{"target": target, "operator": operator, "target-value": value}))

    def configure_node(self) -> Tuple[NodeType, int]:
        """Asks the user for the node type and number of parallel nodes to run on.

        :return: the type of the node and the amount of parallel nodes to run
        """
        node_options = [
            Option(id=node, label=f"{node.name} ({node.cores} cores, {node.ram} GB RAM) @ ${node.price:.2f} per hour")
            for node in available_nodes
        ]

        node = self._logger.prompt_list("Select the optimization node type", node_options)
        parallel_nodes = click.prompt(f"How many nodes should run in parallel ({node.min_nodes}-{node.max_nodes})",
                                      type=click.IntRange(min=node.min_nodes, max=node.max_nodes),
                                      default=node.default_nodes)

        return node, parallel_nodes

    def parse_target(self, target: str) -> str:
        """Parses a target given by the user.

        Converts a target like "Sharpe Ratio" into "TotalPerformance.PortfolioStatistics.SharpeRatio".

        :param target: the target given by the user
        :return: the target in a way it can be passed to the optimizer
        """
        if "." in target:
            return target

        # Turn "SharpeRatio" into "Sharpe Ratio" so the title() call doesn't lowercase the R
        target = re.sub(r"([A-Z])", r" \1", target)

        return f"TotalPerformance.PortfolioStatistics.{target.title().replace(' ', '')}"

    def parse_parameters(self, parameters: List[Tuple[str, float, float, float]]) -> List[OptimizationParameter]:
        """Parses a list of parameters given by the user into a list of parameter objects.

        :param parameters: the parameters given by the user
        :return: the parameters the user gave as OptimizationParameter objects
        """
        parsed_parameters = []

        for name, minimum, maximum, step in parameters:
            parsed_parameters.append(OptimizationParameter(name=name, min=minimum, max=maximum, step=step))

        return parsed_parameters

    def parse_constraints(self, constraints: List[str]) -> List[OptimizationConstraint]:
        """Parses a list of constraints given by the user into a list of constraint objects.

        :param constraints: the constraints given by the user
        :return: the constraints the user gave as OptimizationConstraint objects
        """
        parsed_constraints = []

        for constraint in constraints:
            parts = constraint.rsplit(" ", 2)

            operator = {
                ">": OptimizationConstraintOperator.Greater,
                "<": OptimizationConstraintOperator.Less,
                ">=": OptimizationConstraintOperator.GreaterOrEqual,
                "<=": OptimizationConstraintOperator.LessOrEqual,
                "==": OptimizationConstraintOperator.Equals,
                "!=": OptimizationConstraintOperator.NotEqual
            }[parts[1]]

            parsed_constraints.append(OptimizationConstraint(**{"target": self.parse_target(parts[0]),
                                                                "operator": operator,
                                                                "target-value": float(parts[2])}))

        return parsed_constraints



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from lean.components.api.api_client import *
from lean.models.api import QCDataset


class MarketClient:
    """The MarketClient class contains methods to interact with market/* API endpoints."""

    def __init__(self, api_client: 'APIClient') -> None:
        """Creates a new MarketClient instance.

        :param api_client: the APIClient instance to use when making requests
        """
        self._api = api_client

    def list_datasets(self) -> List[QCDataset]:
        """Get all available datasets.

        :return: all datasets on QuantConnect Datasets
        """
        data = self._api.post("market/data/list")

        return [QCDataset(**dataset) for dataset in data["list"]]



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pathlib import Path
import uuid
import click
from lean.click import LeanCommand, PathParameter
from lean.commands.live.live import get_result, send_command


@click.command(cls=LeanCommand, requires_lean_config=True, requires_docker=True)
@click.argument("project", type=PathParameter(exists=True, file_okay=True, dir_okay=True))
def stop(project: Path) -> None:
    """
    Stop an already running local live trading project.
    """
    command_id = uuid.uuid4().hex

    data = {
        "$type": "QuantConnect.Commands.AlgorithmStatusCommand, QuantConnect.Common",
        "Id": command_id,
        "Status": "Stopped"
    }

    docker_container_name = send_command(project, data)
    get_result(command_id, docker_container_name)




# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lean.models.brokerages.cloud.cloud_brokerage import CloudBrokerage
from lean.models import json_modules
from typing import List

all_cloud_brokerages: List[CloudBrokerage] = []

for json_module in json_modules:
    if "cloud-brokerage" in json_module["type"]:
        all_cloud_brokerages.append(CloudBrokerage(json_module))

[PaperTradingBrokerage] = [
    cloud_brokerage for cloud_brokerage in all_cloud_brokerages if cloud_brokerage._id == "QuantConnectBrokerage"]



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import multiprocessing
import re
from datetime import datetime
from enum import Enum
from typing import List, Any, Optional, Dict, Set, Tuple, Pattern

import click
from dateutil.rrule import rrule, DAILY
from joblib import Parallel, delayed
from pydantic import validator

from lean.click import DateParameter
from lean.container import container
from lean.models.api import QCDataVendor
from lean.models.logger import Option
from lean.models.pydantic import WrappedBaseModel


class OptionResult(WrappedBaseModel):
    """The OptionResult class represents an option's result with an internal value and a display-friendly label."""
    value: Any
    label: str


class DatasetCondition(WrappedBaseModel, abc.ABC):
    def check(self, option_results: Dict[str, OptionResult]) -> bool:
        """Evaluates the condition against a set of options.

        If there is not enough information because of missing options the condition evaluates to True.

        :param option_results: the option id -> option result dictionary to evaluate against
        :return: False if the condition definitely
        """
        raise NotImplementedError()


class DatasetOneOfCondition(DatasetCondition):
    option: str
    values: List[str]

    def check(self, option_results: Dict[str, OptionResult]) -> bool:
        if self.option not in option_results:
            return True
        # TODO: bug? ^ returns true even if option hasn't been resolved

        return option_results[self.option].value in self.values

class DatasetOrCondition(DatasetCondition):
    options: List[DatasetCondition]

    def check(self, option_results: Dict[str, OptionResult]) -> bool:
        # Check each option, if any return true then its true
        for option in self.options:
            if option.check(option_results):
                return True
        
        return False

class DatasetAndCondition(DatasetCondition):
    options: List[DatasetCondition]

    def check(self, option_results: Dict[str, OptionResult]) -> bool:
        # Check each option, if any return false, then its false
        for option in self.options:
            if not option.check(option_results):
                return False

        return True

class DatasetOption(WrappedBaseModel, abc.ABC):
    id: str
    label: str
    description: str
    condition: Optional[DatasetCondition] = None

    @validator("condition", pre=True)
    def parse_condition(cls, value: Optional[Any]) -> Any:
        if value is None or isinstance(value, DatasetCondition):
            return value

        condition_types = {
            "oneof": DatasetOneOfCondition,
            "and": DatasetAndCondition,
            "or": DatasetOrCondition
        }

        conditionType = value["type"].lower()

        # Special AND/OR (Composition of conditions)
        if conditionType == "and" or conditionType == "or" :
            for i in range(0, len(value["options"])):
                option = value["options"][i]

                # Recurse as needed to flush out conditional tree
                value["options"][i] = cls.parse_condition(option)

        return condition_types[conditionType](**value)

    def configure_interactive(self) -> OptionResult:
        """Prompt the user for input to configure this option.

        :return: the parsed result
        """
        raise NotImplementedError()

    def configure_non_interactive(self, user_input: str) -> OptionResult:
        """Parses user input without prompting for anything.

        Raises a ValueError with a descriptive error message if the given input is invalid.

        :param user_input: the input given by the user
        :return: the parsed result
        """
        raise NotImplementedError()

    def get_placeholder(self) -> str:
        """Returns the placeholder of this option.

        :return: the value to show in the place of "value" when documenting this option as "--id <value>"
        """
        raise NotImplementedError()


class DatasetTextOptionTransform(str, Enum):
    Lowercase = "lowercase"
    Uppercase = "uppercase"

    def apply(self, value: str) -> str:
        if self == DatasetTextOptionTransform.Lowercase:
            return value.lower()
        elif self == DatasetTextOptionTransform.Uppercase:
            return value.upper()


class DatasetTextOption(DatasetOption):
    transform: DatasetTextOptionTransform
    multiple: bool = False

    def configure_interactive(self) -> OptionResult:
        prompt = self.label
        if self.multiple:
            prompt += " (comma-separated)"

        user_input = click.prompt(prompt)
        return self.configure_non_interactive(user_input)

    def configure_non_interactive(self, user_input: str) -> OptionResult:
        if len(user_input.strip()) == 0:
            raise ValueError("Value cannot be a blank string")

        if self.multiple:
            parts = [v.strip() for v in user_input.split(",") if v.strip() != ""]
            value = [self.transform.apply(p) for p in parts]
            label = ", ".join(parts)
        else:
            value = self.transform.apply(user_input)
            label = user_input

        return OptionResult(value=value, label=label)

    def get_placeholder(self) -> str:
        if self.multiple:
            return "values"
        else:
            return "value"


class DatasetSelectOption(DatasetOption):
    choices: Dict[str, str]

    def configure_interactive(self) -> OptionResult:
        logger = container.logger()

        keys = list(self.choices.keys())

        if len(keys) <= 5:
            key = logger.prompt_list(self.label, [Option(id=key, label=key) for key in keys])
        else:
            while True:
                user_input = click.prompt(f"{self.label} (example: {min(keys, key=len)})")

                key = next((key for key in keys if key.lower() == user_input.lower()), None)
                if key is not None:
                    break

                logger.info(f"Error: '{user_input}' is not a valid option")

        return OptionResult(value=self.choices[key], label=key)

    def configure_non_interactive(self, user_input: str) -> OptionResult:
        keys = list(self.choices.keys())

        key = next((key for key in keys if key.lower() == user_input.lower()), None)
        if key is None:
            error = f"'{user_input}' is not a valid option"

            if len(keys) <= 5:
                error += f", please choose one of the following: {', '.join(keys)}"
            else:
                error += f", please specify a value like '{min(keys, key=len)}'"

            raise ValueError(error)

        return OptionResult(value=self.choices[key], label=key)

    def get_placeholder(self) -> str:
        keys = list(self.choices.keys())

        if len(keys) <= 5:
            return "|".join(keys)
        else:
            return f"value (example: {min(keys, key=len)})"

class DatasetDateOption(DatasetOption):
    start_end: bool = False

    def configure_interactive(self) -> OptionResult:
        date = click.prompt(f"{self.label} (yyyyMMdd)", type=DateParameter())
        return OptionResult(value=date, label=date.strftime("%Y-%m-%d"))

    def configure_non_interactive(self, user_input: str) -> OptionResult:
        for date_format in ["%Y%m%d", "%Y-%m-%d"]:
            try:
                date = datetime.strptime(user_input, date_format)
                return OptionResult(value=date, label=date.strftime("%Y-%m-%d"))
            except ValueError:
                pass

        raise ValueError(f"'{user_input}' does not match the yyyyMMdd format")

    def get_placeholder(self) -> str:
        return "yyyyMMdd"


class DatasetPathTemplates(WrappedBaseModel):
    all: List[str] = []
    latest: List[str] = []


class DatasetPath(WrappedBaseModel):
    condition: Optional[DatasetCondition] = None
    templates: DatasetPathTemplates

    @validator("condition", pre=True)
    def parse_condition(cls, value: Optional[Any]) -> Any:
        if value is None or isinstance(value, DatasetCondition):
            return value

        condition_types = {
            "oneof": DatasetOneOfCondition,
            "and": DatasetAndCondition,
            "or": DatasetOrCondition
        }

        conditionType = value["type"].lower()

        # Special AND/OR (Composition of conditions)
        if conditionType == "and" or conditionType == "or" :
            for i in range(0, len(value["options"])):
                option = value["options"][i]

                # Recurse as need to flush out conditional tree
                value["options"][i] = cls.parse_condition(option)

        return condition_types[conditionType](**value)


class Dataset(WrappedBaseModel):
    name: str
    vendor: str
    categories: List[str]
    options: List[DatasetOption]
    paths: List[DatasetPath]
    requires_security_master: bool

    @validator("options", pre=True)
    def parse_options(cls, values: List[Any]) -> List[Any]:
        option_types = {
            "text": DatasetTextOption,
            "select": DatasetSelectOption,
            "date": DatasetDateOption
        }

        options = []
        for option in values:
            if isinstance(option, DatasetOption):
                options.append(option)

            # TODO: This is a hack around, does not respect option conditions for start-end
            elif option["type"] == "start-end":
                description_suffix = ""
                required_resolutions = ["tick", "second", "minute", "minute/second/tick"]

                resolution = next((o for o in options if o.id == "resolution"), None)
                if resolution is not None and isinstance(resolution, DatasetSelectOption):
                    if len(set(resolution.choices.values()) - set(required_resolutions)) > 0:
                        description_suffix = " (tick, second and minute resolutions only)"

                options.extend([
                    DatasetDateOption(
                        id="start",
                        label="Start date",
                        description="The inclusive end date of the data that you want to download" + description_suffix,
                        condition=DatasetOneOfCondition(option="resolution", values=required_resolutions),
                        start_end=True
                    ),
                    DatasetDateOption(
                        id="end",
                        label="End date",
                        description="The inclusive end date of the data that you want to download" + description_suffix,
                        condition=DatasetOneOfCondition(option="resolution", values=required_resolutions),
                        start_end=True
                    )
                ])
            else:
                options.append(option_types[option["type"]](**option))

        return options


class DataFile(WrappedBaseModel):
    file: str
    vendor: QCDataVendor


class DataFileGroup(WrappedBaseModel, abc.ABC):
    prefix: str

    def get_valid_files(self, files_with_prefix: Optional[List[str]]) -> Set[str]:
        raise NotImplementedError()


class DataFileAllGroup(DataFileGroup):
    possible_files: Set[str]

    def get_valid_files(self, files_with_prefix: Optional[List[str]]) -> Set[str]:
        if files_with_prefix is not None:
            return self.possible_files.intersection(files_with_prefix)

        return self.possible_files


class DataFileLatestGroup(DataFileGroup):
    regex: Pattern

    def get_valid_files(self, files_with_prefix: Optional[List[str]]) -> Set[str]:
        if files_with_prefix is not None:
            matching_files = [file for file in files_with_prefix if self.regex.match(file) is not None]
            if len(matching_files) > 0:
                return {sorted(matching_files)[-1]}

        return set()


class Product(WrappedBaseModel):
    dataset: Dataset
    option_results: Dict[str, OptionResult]

    def get_data_files(self) -> List[str]:
        """Returns all data files for the given product configuration.

        :return: the list of files that need to be downloaded for this product
        """
        groups = []
        variables = {option_id: result.value for option_id, result in self.option_results.items()}

        multiple_option = next((o for o in self.dataset.options if isinstance(o, DatasetTextOption) and o.multiple),
                               None)
        if multiple_option is not None and multiple_option.id in self.option_results:
            result = self.option_results[multiple_option.id]

            for index in range(len(result.value)):
                groups.extend(self._get_data_file_groups({
                    **variables,
                    multiple_option.id: result.value[index]
                }))
        else:
            groups.extend(self._get_data_file_groups(variables))

        prefixes = set(group.prefix for group in groups)
        prefixes_to_files = {}

        parallel = Parallel(n_jobs=max(1, multiprocessing.cpu_count() - 1), backend="threading")
        for prefix, files_with_prefix in parallel(delayed(self._list_files)(prefix) for prefix in prefixes):
            prefixes_to_files[prefix] = files_with_prefix

        data_files = set()
        for group in groups:
            data_files.update(group.get_valid_files(prefixes_to_files[group.prefix]))

        return sorted(list(data_files))

    def _get_data_file_groups(self, variables: Dict[str, Any]) -> List[DataFileGroup]:
        groups = []

        for path in self.dataset.paths:
            if path.condition is None or path.condition.check(self.option_results):
                path_to_use = path
                break
        else:
            raise RuntimeError(f"No eligible path templates found")

        for template in path_to_use.templates.all:
            has_start_end = any(isinstance(o, DatasetDateOption) and o.start_end for o in self.dataset.options)
            start = variables.get("start", None)
            end = variables.get("end", None)

            possible_files = set()

            if has_start_end and start is not None and end is not None:
                variables_to_use = {**variables}
                for date in rrule(DAILY, dtstart=start, until=end):
                    variables_to_use["date"] = date
                    variables_to_use["year"] = date.strftime("%Y")
                    variables_to_use["month"] = date.strftime("%m")
                    variables_to_use["day"] = date.strftime("%d")
                    possible_files.add(self._render_template(template, variables_to_use))
            else:
                possible_files.add(self._render_template(template, variables))

            prefix = self._get_common_prefix(list(possible_files))

            groups.append(DataFileAllGroup(prefix=prefix, possible_files=possible_files))

        for regex_template in path_to_use.templates.latest:
            rendered_regex = self._render_template(regex_template, variables)

            prefix = re.split(r"[\\[\]()]", rendered_regex)[0]
            compiled_regex = re.compile(rendered_regex)

            groups.append(DataFileLatestGroup(prefix=prefix, regex=compiled_regex))

        return groups

    def _list_files(self, prefix: str) -> Tuple[str, Optional[List[str]]]:
        if len(prefix.split("/")) < 3:
            # Cannot get cloud directory listing less than 3 levels deep
            return prefix, None
        else:
            return prefix, container.api_client().data.list_files(prefix)

    def _get_common_prefix(self, values: List[str]) -> str:
        """Finds the common prefix in a list of strings.

        :param values: the strings to find the common prefix of
        :return: the common prefix of the given strings
        """
        shortest_value = min(values, key=len)

        for index, character in enumerate(shortest_value):
            for value in values:
                if value[index] != character:
                    return shortest_value[:index]

        return shortest_value

    def _render_template(self, template: str, variables: Dict[str, str]) -> str:
        """Renders a template string with variables.

        Variables can be referenced in the template by putting accolades around them.
        Example template: "path/to/{date}.zip" (date is a referenced variable here)

        :param template: the template to render
        :param variables: the variables that are accessible to the template
        :return: the rendered template
        """
        for key, value in variables.items():
            if isinstance(value, datetime):
                value = value.strftime("%Y%m%d")

            template = template.replace("{" + key + "}", str(value))

        return template



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webbrowser
from typing import Optional
import click
from lean.click import LeanCommand
from lean.container import container
from pathlib import Path
from lean.models.errors import RequestFailedError

@click.command(cls=LeanCommand)
@click.argument("project", type=str)
@click.option("--name", type=str, help="The name of the backtest (a random one is generated if not specified)")
@click.option("--push",
              is_flag=True,
              default=False,
              help="Push local modifications to the cloud before running the backtest")
@click.option("--open", "open_browser",
              is_flag=True,
              default=False,
              help="Automatically open the results in the browser when the backtest is finished")
def backtest(project: str, name: Optional[str], push: bool, open_browser: bool) -> None:
    """Backtest a project in the cloud.

    PROJECT must be the name or id of the project to run a backtest for.

    If the project that has to be backtested has been pulled to the local drive
    with `lean cloud pull` it is possible to use the --push option to push local
    modifications to the cloud before running the backtest.
    """
    logger = container.logger()

    cloud_project_manager = container.cloud_project_manager()
    try:
        cloud_project = cloud_project_manager.get_cloud_project(project, push)
    except RuntimeError as e:
        if cloud_project_manager._project_config_manager.try_get_project_config(Path.cwd() / project,
                                                                        cloud_project_manager._path_manager):
            error_message = f'No project with the given name or id "{project}" found in your cloud projects.'
            error_message += f" Please use `lean cloud backtest --push {project}` to backtest in cloud."
        else:
            error_message = f'No project with the given name or id "{project}" found in your cloud or local projects.'
        raise RuntimeError(error_message)

    if name is None:
        name = container.name_generator().generate_name()

    cloud_runner = container.cloud_runner()
    finished_backtest = cloud_runner.run_backtest(cloud_project, name)

    if finished_backtest.error is None and finished_backtest.stacktrace is None:
        logger.info(finished_backtest.get_statistics_table())

    logger.info(f"Backtest id: {finished_backtest.backtestId}")
    logger.info(f"Backtest name: {finished_backtest.name}")
    logger.info(f"Backtest url: {finished_backtest.get_url()}")

    if finished_backtest.error is not None or finished_backtest.stacktrace is not None:
        error = finished_backtest.stacktrace or finished_backtest.error
        error = error.strip()

        logger.error("An error occurred during this backtest:")
        logger.error(error)

        # Don't open the results in the browser if the error happened during initialization
        # In the browser the logs won't show these errors, you'll just get empty charts and empty logs
        if error.startswith("During the algorithm initialization, the following exception has occurred:"):
            open_browser = False

    if open_browser:
        webbrowser.open(finished_backtest.get_url())



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Optional

import click

from lean.click import DateParameter, LeanCommand
from lean.constants import DEFAULT_ENGINE_IMAGE
from lean.container import container


@click.command(cls=LeanCommand, requires_lean_config=True, requires_docker=True)
@click.option("--start",
              type=DateParameter(),
              required=True,
              help="Start date for the data to generate in yyyyMMdd format")
@click.option("--end",
              type=DateParameter(),
              default=datetime.today().strftime("%Y%m%d"),
              help="End date for the data to generate in yyyyMMdd format (defaults to today)")
@click.option("--symbol-count",
              type=click.IntRange(min=0),
              required=True,
              help="The number of symbols to generate data for")
@click.option("--security-type",
              type=click.Choice(["Equity", "Forex", "Cfd", "Future", "Crypto", "Option"], case_sensitive=False),
              default="Equity",
              help="The security type to generate data for (defaults to Equity)")
@click.option("--resolution",
              type=click.Choice(["Tick", "Second", "Minute", "Hour", "Daily"], case_sensitive=False),
              default="Minute",
              help="The resolution of the generated data (defaults to Minute)")
@click.option("--data-density",
              type=click.Choice(["Dense", "Sparse", "VerySparse"], case_sensitive=False),
              default="Dense",
              help="The density of the generated data (defaults to Dense)")
@click.option("--include-coarse",
              type=bool,
              default=True,
              help="Whether coarse universe data should be generated for Equity data (defaults to True)")
@click.option("--market",
              type=str,
              default="",
              help="The market to generate data for (defaults to standard market for the security type)")
@click.option("--image",
              type=str,
              help=f"The LEAN engine image to use (defaults to {DEFAULT_ENGINE_IMAGE})")
@click.option("--update",
              is_flag=True,
              default=False,
              help="Pull the LEAN engine image before running the generator")
def generate(start: datetime,
             end: datetime,
             symbol_count: int,
             security_type: str,
             resolution: str,
             data_density: str,
             include_coarse: bool,
             market: str,
             image: Optional[str],
             update: bool) -> None:
    """Generate random market data.

    \b
    This uses the random data generator in LEAN to generate realistic market data using a Brownian motion model.
    This generator supports the following security types, tick types and resolutions:
    | Security type | Generated tick types | Supported resolutions                |
    | ------------- | -------------------- | ------------------------------------ |
    | Equity        | Trade                | Tick, Second, Minute, Hour and Daily |
    | Forex         | Quote                | Tick, Second, Minute, Hour and Daily |
    | CFD           | Quote                | Tick, Second, Minute, Hour and Daily |
    | Future        | Trade and Quote      | Tick, Second, Minute, Hour and Daily |
    | Crypto        | Trade and Quote      | Tick, Second, Minute, Hour and Daily |
    | Option        | Trade and Quote      | Minute                               |

    \b
    The following data densities are available:
    - Dense: at least one data point per resolution step.
    - Sparse: at least one data point per 5 resolution steps.
    - VerySparse: at least one data point per 50 resolution steps.

    \b
    Example which generates minute data for 100 equity symbols since 2015-01-01:
    $ lean data generate --start=20150101 --symbol-count=100

    \b
    Example which generates daily data for 100 crypto symbols since 2015-01-01:
    $ lean data generate --start=20150101 --symbol-count=100 --security-type=Crypto --resolution=Daily

    By default the official LEAN engine image is used.
    You can override this using the --image option.
    Alternatively you can set the default engine image for all commands using `lean config set engine-image <image>`.
    """
    lean_config_manager = container.lean_config_manager()
    data_dir = lean_config_manager.get_data_directory()

    run_options = {
        "entrypoint": ["dotnet", "QuantConnect.ToolBox.dll",
                       "--destination-dir", "/Lean/Data",
                       "--app", "randomdatagenerator",
                       "--start", start.strftime("%Y%m%d"),
                       "--end", end.strftime("%Y%m%d"),
                       "--symbol-count", str(symbol_count),
                       "--security-type", security_type,
                       "--resolution", resolution,
                       "--data-density", data_density,
                       "--include-coarse", str(include_coarse).lower(),
                       "--market", market.lower()],
        "volumes": {
            str(data_dir): {
                "bind": "/Lean/Data",
                "mode": "rw"
            }
        }
    }

    engine_image = container.cli_config_manager().get_engine_image(image)

    container.update_manager().pull_docker_image_if_necessary(engine_image, update)

    success = container.docker_manager().run_image(engine_image, **run_options)
    if not success:
        raise RuntimeError(
            "Something went wrong while running the random data generator, see the logs above for more information")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform
import sys
from typing import Any, List, Optional

import click
import maskpass
from rich.console import Console
from rich.progress import BarColumn, Progress, TextColumn

from lean.models.logger import Option


class Logger:
    """The Logger class handles all output printing."""

    def __init__(self) -> None:
        """Creates a new Logger instance."""
        self._console = Console(markup=False, highlight=False, emoji=False, width=None)
        self.debug_logging_enabled = False

    def debug(self, message: Any) -> None:
        """Logs a debug message if debug logging is enabled.

        :param message: the message to log
        """
        if self.debug_logging_enabled:
            self._console.print(message, style="grey50")

    def info(self, message: Any) -> None:
        """Logs an info message.

        :param message: the message to log
        """
        self._console.print(message)

    def warn(self, message: Any) -> None:
        """Logs a warning message.

        :param message: the message to log
        """
        self._console.print(message, style="yellow")

    def error(self, message: Any) -> None:
        """Logs an error message.

        :param message: the message to log
        """
        self._console.print(message, style="red")

    def progress(self, prefix: str = "", suffix: str = "{task.percentage:0.0f}%") -> Progress:
        """Creates a Progress instance.

        :param prefix: the text to show before the bar (defaults to a blank string)
        :param suffix: the text to show after the bar (defaults to the task's percentage)
        :return: a Progress instance which can be used to display progress bars
        """
        progress = Progress(TextColumn(prefix), BarColumn(), TextColumn(suffix), console=self._console)
        progress.start()
        return progress

    def prompt_list(self, text: str, options: List[Option], default: Optional[str] = None, multiple: bool = False) -> Any:
        """Asks the user to select an option from a list of possible options.

        The user will not be prompted for input if there is only a single option.

        :param text: the text to display before prompting
        :param options: the available options
        :param default: the default value if no input is given
        :return: the chosen option's id
        """
        def validate_option(input_value: Any):
            try:
                index = int(input_value)
                if 0 < index <= len(options):
                    return options[index - 1].id
            except ValueError:
                option = next((option for option in options if option.label == input_value), None)
                if option is not None:
                    return option.id

            self.info("Please enter the number or label of an option")

        if len(options) == 1:
            self.info(f"{text}: {options[0].label}")
            return options[0].id

        self.info(f"{text}:")
        for i, option in enumerate(options):
            self.info(f"{i + 1}) {option.label}")

        while True:
            if not multiple:
                user_input = click.prompt("Enter an option", type=str, default=default, show_default=True)
                user_selected_value = validate_option(user_input)
                if user_selected_value is not None:
                    return user_selected_value
            else:
                user_selected_values = []
                user_inputs = click.prompt("To enter multiple options, separate them with comma.", type=str, default=default, show_default=True)
                user_inputs = str(user_inputs).strip(",").split(",")
                expected_outputs = len(user_inputs)
                for user_input in user_inputs:
                    user_selected_value = validate_option(user_input)
                    if user_selected_value is not None:
                        user_selected_values.append(user_selected_value)
                if len(user_selected_values) == expected_outputs:
                    return user_selected_values

    def prompt_password(self, text: str, default: Optional[str] = None) -> str:
        """Asks the user for a string value while masking the given input.

        :param text: the text to display before prompting
        :param default: the default value if no input is given
        :return: the given input
        """
        if default is not None:
            text = f"{text} [{'*' * len(default)}]"

        # Masking does not work properly in WSL2 and when the input is not coming from a keyboard
        if "microsoft" in platform.uname().release.lower() or not sys.stdin.isatty():
            return click.prompt(text, default=default, show_default=False)

        while True:
            user_input = maskpass.askpass(f"{text}: ")

            if len(user_input) == 0 and default is not None:
                return default

            if len(user_input) > 0:
                return user_input



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pathlib import Path
from typing import Optional
import uuid
import click
from lean.click import LeanCommand, PathParameter
from lean.commands.live.live import get_result, send_command
from lean.components.util.click_custom_parameters import DECIMAL

@click.command(cls=LeanCommand, requires_lean_config=True, requires_docker=True)
@click.argument("project", type=PathParameter(exists=True, file_okay=True, dir_okay=True))
@click.option("--ticker", type=str, required=True, help="The ticker of the symbol to add")
@click.option("--market", type=str, required=True, help="The market of the symbol to add")
@click.option("--security-type", type=str, required=True, help="The security type of the symbol to add")
@click.option("--resolution",
              type=str,
              default="Minute",
              help="The resolution of the symbol to add")
@click.option("--fill-data-forward",
              is_flag=True,
              default=True,
              help="The fill forward behavior, true to fill forward, false otherwise - defaults to true")
@click.option("--leverage",
              type=DECIMAL,
              default=0.0,
              help="The leverage for the security, defaults to 2 for equity, 50 for forex, and 1 for everything else")
@click.option("--extended-market-hours",
              is_flag=True,
              default=False,
              help="The extended market hours flag, true to allow pre/post market data, false for only in market data")
def add_security(project: Path,
                 ticker: str,
                 market: str,
                 security_type: str,
                 resolution: Optional[str],
                 fill_data_forward: Optional[bool],
                 leverage: Optional[DECIMAL],
                 extended_market_hours: Optional[bool]) -> None:
    """
    Represents a command to add a security to the algorithm.
    """
    command_id = uuid.uuid4().hex

    data = {
        "$type": "QuantConnect.Commands.AddSecurityCommand, QuantConnect.Common",
        "Id": command_id,
        "Symbol": ticker,
        "Market": market,
        "SecurityType": security_type,
        "Resolution": resolution,
        "FillDataForward": fill_data_forward,
        "Leverage": leverage,
        "ExtendedMarketHours": extended_market_hours
    }

    docker_container_name = send_command(project, data)
    get_result(command_id, docker_container_name)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
import re
from typing import Any, Dict, List
import click
import abc
from lean.components.util.logger import Logger
from lean.click import PathParameter


class BaseCondition(abc.ABC):
    """Base condition class extended to all types of conditions"""

    def __init__(self, condition_object: Dict[str, str]):
        self._type: str = condition_object["type"]
        self._pattern: str = str(condition_object["pattern"])
        self._dependent_config_id: str = condition_object["dependent-config-id"]

    def factory(condition_object: Dict[str, str]) -> 'BaseCondition':
        """Creates an instance of the child classes.

        :param condition_object: the json object dict with condition info
        :raises ValueError: When the wrong condition type is provided
        :return: An instance of BaseCondition
        """
        if condition_object["type"] == "regex":
            return RegexCondition(condition_object)
        elif condition_object["type"] == "exact-match":
            return ExactMatchCondition(condition_object)
        else:
            raise ValueError(
                f'Undefined condition type {condition_object["type"]}')

    @abc.abstractmethod
    def check(self, target_value: str) -> bool:
        """validates the condition against the provided values

        :param target_value: value to validate the condition against
        :return: True if the condition is valid otherwise False
        """
        raise NotImplementedError()


class ExactMatchCondition(BaseCondition):
    """This class is used when the condition needs to be evaluated with equality"""

    def check(self, target_value: str) -> bool:
        """validates the condition against the provided values

        :param target_value: value to validate the condition against
        :return: True if the condition is valid otherwise False
        """
        return self._pattern.casefold() == target_value.casefold()


class RegexCondition(BaseCondition):
    """This class is used when the condition needs to be evaluated using regex"""

    def check(self, target_value: str) -> bool:
        """validates the condition against the provided values

        :param target_value: value to validate the condition against
        :return: True if the condition is valid otherwise False
        """
        return len(re.findall(self._pattern, target_value, re.I)) > 0


class ConditionalValueOption():
    """This class is used when mutliple values needs to be evaluated based on conditions."""

    def __init__(self, option_object: Dict[str, Any]):
        self._value: str = option_object["value"]
        self._condition: BaseCondition = BaseCondition.factory(
            option_object["condition"])


class Configuration(abc.ABC):
    """Base configuration class extended to all types of configurations"""

    def __init__(self, config_json_object):
        self._id: str = config_json_object["id"]
        self._config_type: str = config_json_object["type"]
        self._value: str = config_json_object["value"]
        self._is_cloud_property: bool = "cloud-id" in config_json_object
        self._is_required_from_user = False
        self._is_type_configurations_env: bool = type(
            self) is ConfigurationsEnvConfiguration
        self._is_type_trading_env: bool = type(self) is TradingEnvConfiguration
        self.is_type_organization_id: bool = type(
            self) is OrganzationIdConfiguration
        self._log_message: str = ""
        if "log-message" in config_json_object.keys():
            self._log_message = config_json_object["log-message"]
        if "filters" in config_json_object.keys():
            self._filter = Filter(config_json_object["filters"])
        else:
            self._filter = Filter([])

    def factory(config_json_object) -> 'Configuration':
        """Creates an instance of the child classes.

        :param config_json_object: the json object dict with configuration info
        :raises ValueError: When the wrong configuration type is provided.
        :return: An instance of Configuration.
        """

        if config_json_object["type"] in ["info", "configurations-env"]:
            return InfoConfiguration.factory(config_json_object)
        elif config_json_object["type"] in ["input", "internal-input"]:
            return UserInputConfiguration.factory(config_json_object)
        elif config_json_object["type"] == "filter-env":
            return BrokerageEnvConfiguration.factory(config_json_object)
        elif config_json_object["type"] == "trading-env":
            return TradingEnvConfiguration.factory(config_json_object)
        elif config_json_object["type"] == "organization-id":
            return OrganzationIdConfiguration(config_json_object)
        else:
            raise ValueError(
                f'Undefined input method type {config_json_object["type"]}')


class Filter():
    """This class handles the conditional filters added to configurations.
    """

    def __init__(self, filter_conditions):
        self._conditions: List[BaseCondition] = [BaseCondition.factory(
            condition["condition"]) for condition in filter_conditions]


class InfoConfiguration(Configuration):
    """Configuration class used for informational configurations.

    Doesn't support user prompt inputs.
    Values of this configuration isn't persistently saved in the Lean configuration.
    """

    def __init__(self, config_json_object):
        super().__init__(config_json_object)
        self._is_required_from_user = False

    def factory(config_json_object) -> 'InfoConfiguration':
        """Creates an instance of the child classes.

        :param config_json_object: the json object dict with configuration info
        :return: An instance of InfoConfiguration.
        """
        if config_json_object["type"] == "configurations-env":
            return ConfigurationsEnvConfiguration(config_json_object)
        else:
            return InfoConfiguration(config_json_object)


class ConfigurationsEnvConfiguration(InfoConfiguration):
    """Configuration class used for environment properties.

    Doesn't support user prompt inputs.
    Values of this configuration isn't persistently saved in the Lean configuration.
    """

    def __init__(self, config_json_object):
        super().__init__(config_json_object)
        self._env_and_values = {
            env_obj["name"]: env_obj["value"] for env_obj in self._value}


class UserInputConfiguration(Configuration, abc.ABC):
    """Base class extended to all confiugration class than store values in Lean config.

    Values are expected from the user via prompts.
    Values of this configuration is persistently saved in the Lean configuration,
    until specified explicitly.
    """

    def __init__(self, config_json_object):
        super().__init__(config_json_object)
        self._is_required_from_user = True
        self._input_method = self._prompt_info = self._help = ""
        self._input_default = self._cloud_id = None
        if "input-method" in config_json_object.keys():
            self._input_method = config_json_object["input-method"]
        if "prompt-info" in config_json_object.keys():
            self._prompt_info = config_json_object["prompt-info"]
        if "help" in config_json_object.keys():
            self._help = config_json_object["help"]
        if "input-default" in config_json_object.keys():
            self._input_default = config_json_object["input-default"]
        if "cloud-id" in config_json_object.keys():
            self._cloud_id = config_json_object["cloud-id"]

    @abc.abstractmethod
    def AskUserForInput(self, default_value: Any, logger: Logger):
        """Prompts user to provide input while validating the type of input
        against the expected type

        :param default_value: The default to prompt to the user.
        :param logger: The instance of logger class.
        :return: The value provided by the user.
        """
        return NotImplemented()

    def factory(config_json_object) -> 'UserInputConfiguration':
        """Creates an instance of the child classes.

        :param config_json_object: the json object dict with configuration info
        :return: An instance of UserInputConfiguration.
        """
        # NOTE: Check "Type" before "Input-method"
        if config_json_object["type"] == "internal-input":
            return InternalInputUserInput(config_json_object)
        if config_json_object["input-method"] == "prompt":
            return PromptUserInput(config_json_object)
        elif config_json_object["input-method"] == "choice":
            return ChoiceUserInput(config_json_object)
        elif config_json_object["input-method"] == "confirm":
            return ConfirmUserInput(config_json_object)
        elif config_json_object["input-method"] == "prompt-password":
            return PromptPasswordUserInput(config_json_object)
        elif config_json_object["input-method"] == "path-parameter":
            return PathParameterUserInput(config_json_object)


class InternalInputUserInput(UserInputConfiguration):
    """This class is used when configuratios are needed by LEAN config but the values
        are derived from other dependent configurations and not from user input."""

    def __init__(self, config_json_object):
        super().__init__(config_json_object)
        self._is_conditional: bool = False
        value_options: List[ConditionalValueOption] = []
        if "value-options" in config_json_object.keys():
            value_options = [ConditionalValueOption(
                value_option) for value_option in config_json_object["value-options"]]
            self._is_conditional = True
        self._value_options = value_options

    def AskUserForInput(self, default_value, logger: Logger):
        """Prompts user to provide input while validating the type of input
        against the expected type

        :param default_value: The default to prompt to the user.
        :param logger: The instance of logger class.
        :return: The value provided by the user.
        """
        raise ValueError(
            f'user input not allowed with {self.__class__.__name__}')


class PromptUserInput(UserInputConfiguration):
    map_to_types = {
        "string": str,
        "boolean": bool,
        "integer": int
    }

    def __init__(self, config_json_object):
        super().__init__(config_json_object)
        self._input_type: str = "string"
        if "input-type" in config_json_object.keys():
            self._input_type = config_json_object["input-type"]

    def AskUserForInput(self, default_value, logger: Logger):
        """Prompts user to provide input while validating the type of input
        against the expected type

        :param default_value: The default to prompt to the user.
        :param logger: The instance of logger class.
        :return: The value provided by the user.
        """
        return click.prompt(self._prompt_info, default_value, type=self.get_input_type())

    def get_input_type(self):
        return self.map_to_types.get(self._input_type, self._input_type)


class ChoiceUserInput(UserInputConfiguration):
    def __init__(self, config_json_object):
        super().__init__(config_json_object)
        self._choices: List[str] = []
        if "input-choices" in config_json_object.keys():
            self._choices = config_json_object["input-choices"]

    def AskUserForInput(self, default_value, logger: Logger):
        """Prompts user to provide input while validating the type of input
        against the expected type

        :param default_value: The default to prompt to the user.
        :param logger: The instance of logger class.
        :return: The value provided by the user.
        """
        return click.prompt(
            self._prompt_info,
            default_value,
            type=click.Choice(self._choices, case_sensitive=False)
        )


class PathParameterUserInput(UserInputConfiguration):
    def __init__(self, config_json_object):
        super().__init__(config_json_object)

    def AskUserForInput(self, default_value, logger: Logger):
        """Prompts user to provide input while validating the type of input
        against the expected type

        :param default_value: The default to prompt to the user.
        :param logger: The instance of logger class.
        :return: The value provided by the user.
        """

        default_binary = None
        if default_value is not None:
            default_binary = Path(default_value)
        elif self._input_default is not None and Path(self._input_default).is_file():
            default_binary = Path(self._input_default)
        else:
            default_binary = ""
        value = click.prompt(self._prompt_info,
                             default=default_binary,
                             type=PathParameter(
                                 exists=False, file_okay=True, dir_okay=False)
                             )
        return value


class ConfirmUserInput(UserInputConfiguration):
    def __init__(self, config_json_object):
        super().__init__(config_json_object)

    def AskUserForInput(self, default_value, logger: Logger):
        """Prompts user to provide input while validating the type of input
        against the expected type

        :param default_value: The default to prompt to the user.
        :param logger: The instance of logger class.
        :return: The value provided by the user.
        """
        return click.prompt(self._prompt_info, default_value, type=bool)


class PromptPasswordUserInput(UserInputConfiguration):
    def __init__(self, config_json_object):
        super().__init__(config_json_object)

    def AskUserForInput(self, default_value, logger: Logger):
        """Prompts user to provide input while validating the type of input
        against the expected type

        :param default_value: The default to prompt to the user.
        :param logger: The instance of logger class.
        :return: The value provided by the user.
        """
        return logger.prompt_password(self._prompt_info, default_value)


class OrganzationIdConfiguration(PromptUserInput):
    """This class is used for job-organzation-id configurations"""

    def __init__(self, config_json_object):
        super().__init__(config_json_object)


class BrokerageEnvConfiguration(PromptUserInput, ChoiceUserInput, ConfirmUserInput):
    """This class is base class extended by all classes that needs to add value to user filters"""

    def __init__(self, config_json_object):
        super().__init__(config_json_object)

    def factory(config_json_object) -> 'BrokerageEnvConfiguration':
        """Creates an instance of the child classes.

        :param config_json_object: the json object dict with configuration info
        :return: An instance of BrokerageEnvConfiguration
        """
        if config_json_object["type"] == "filter-env":
            return FilterEnvConfiguration(config_json_object)
        else:
            raise ValueError(
                f'Undefined input method type {config_json_object["type"]}')

    def AskUserForInput(self, default_value, logger: Logger):
        """Prompts user to provide input while validating the type of input
        against the expected type

        :param default_value: The default to prompt to the user.
        :param logger: The instance of logger class.
        :return: The value provided by the user.
        """
        if self._input_method == "confirm":
            return ConfirmUserInput.AskUserForInput(self, default_value, logger)
        elif self._input_method == "choice":
            return ChoiceUserInput.AskUserForInput(self, default_value, logger)
        elif self._input_method == "prompt":
            return PromptUserInput.AskUserForInput(self, default_value, logger)
        else:
            raise ValueError(
                f"Undefined input method type {self._input_method}")


class TradingEnvConfiguration(PromptUserInput, ChoiceUserInput, ConfirmUserInput):
    """This class adds trading-mode/envirionment based user filters.
    
    Normalizes the value of envrionment values(live/paper) for cloud live. 
    """

    def __init__(self, config_json_object):
        super().__init__(config_json_object)
    
    def factory(config_json_object) -> 'TradingEnvConfiguration':
        """Creates an instance of the child classes.

        :param config_json_object: the json object dict with configuration info
        :return: An instance of TradingEnvConfiguration.
        """
        if config_json_object["type"] == "trading-env":
            return TradingEnvConfiguration(config_json_object)
        else:
            raise ValueError(
                f'Undefined input method type {config_json_object["type"]}')

    def AskUserForInput(self, default_value, logger: Logger):
        """Prompts user to provide input while validating the type of input
        against the expected type

        :param default_value: The default to prompt to the user.
        :param logger: The instance of logger class.
        :return: The value provided by the user.
        """
        # NOTE: trading envrionment config should not use old boolean value as default
        if type(default_value) == bool:
            default_value = "paper" if default_value else "live"
        if self._input_method == "confirm":
            raise ValueError(
                f'input method -- {self._input_method} is not allowed with {self.__class__.__name__}')
        else:
            return BrokerageEnvConfiguration.AskUserForInput(self, default_value, logger)


class FilterEnvConfiguration(BrokerageEnvConfiguration):
    """This class adds extra filters to user filters."""

    def __init__(self, config_json_object):
        super().__init__(config_json_object)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import json
import os
import platform
import shutil
import signal
import subprocess
import sys
import threading
import time
import types
from pathlib import Path
from typing import Optional, Set, Any, Dict
from subprocess import Popen, PIPE
import docker
from dateutil.parser import isoparse
from docker.errors import APIError
from docker.models.containers import Container
from docker.types import Mount

from lean.components.util.logger import Logger
from lean.components.util.platform_manager import PlatformManager
from lean.components.util.temp_manager import TempManager
from lean.constants import SITE_PACKAGES_VOLUME_LIMIT, \
    DOCKER_NETWORK, CUSTOM_FOUNDATION, CUSTOM_RESEARCH, CUSTOM_ENGINE

from lean.models.docker import DockerImage
from lean.models.errors import MoreInfoError
from lean.components.util.custom_json_encoder import DecimalEncoder

class DockerManager:
    """The DockerManager contains methods to manage and run Docker images."""

    def __init__(self, logger: Logger, temp_manager: TempManager, platform_manager: PlatformManager) -> None:
        """Creates a new DockerManager instance.

        :param logger: the logger to use when printing messages
        :param temp_manager: the TempManager instance used when creating temporary directories
        :param platform_manager: the PlatformManager used when checking which operating system is in use
        """
        self._logger = logger
        self._temp_manager = temp_manager
        self._platform_manager = platform_manager

    def pull_image(self, image: DockerImage) -> None:
        """Pulls a Docker image.

        :param image: the image to pull
        """
        if image.name == CUSTOM_RESEARCH or image.name == CUSTOM_ENGINE or image.name == CUSTOM_FOUNDATION:
            self._logger.info(f"Skip pulling local image {image}...")
            return

        self._logger.info(f"Pulling {image}...")
        # We cannot really use docker_client.images.pull() here as it doesn't let us log the progress
        # Downloading multiple gigabytes without showing progress does not provide good developer experience
        # Since the pull command is the same on Windows, macOS and Linux we can safely use a system call
        if shutil.which("docker") is not None:
            process = subprocess.run(["docker", "image", "pull", str(image)])
            if process.returncode != 0:
                raise RuntimeError(
                    f"Something went wrong while pulling {image}, see the logs above for more information")
        else:
            self._get_docker_client().images.pull(image.name, image.tag)

    def run_image(self, image: DockerImage, **kwargs) -> bool:
        """Runs a Docker image. If the image is not available locally it will be pulled first.

        See https://docker-py.readthedocs.io/en/stable/containers.html for all the supported kwargs.

        If kwargs contains an "on_output" property, it is removed before passing it on to docker.containers.run
        and the given lambda is ran whenever the Docker container prints anything.

        If kwargs contains an "format_output" property, it is removed before passing it on to docker.containers.run
        and the given lambda is ran after the Docker container completes running.

        If kwargs contains a "commands" property, it is removed before passing it on to docker.containers.run
        and the Docker container is configured to run the given commands.
        This property causes the "entrypoint" property to be overwritten if it exists.

        If kwargs sets "detach" to True, the method returns as soon as the container starts.
        If this is not the case, the method is blocking and runs until the container exits.

        :param image: the image to run
        :param kwargs: the kwargs to forward to docker.containers.run
        :return: True if the command in the container exited successfully, False if not
        """
        if not self.image_installed(image):
            self.pull_image(image)

        on_output = kwargs.pop("on_output", lambda chunk: None)
        format_output = kwargs.pop("format_output", lambda chunk: None)
        commands = kwargs.pop("commands", None)

        if commands is not None:
            shell_script_commands = ["#!/usr/bin/env bash", "set -e"]
            if self._logger.debug_logging_enabled:
                shell_script_commands.append("set -x")
            shell_script_commands += commands

            shell_script_path = self._temp_manager.create_temporary_directory() / "lean-cli-start.sh"
            with shell_script_path.open("w+", encoding="utf-8", newline="\n") as file:
                file.write("\n".join(shell_script_commands) + "\n")

            if "mounts" not in kwargs:
                kwargs["mounts"] = []

            kwargs["mounts"].append(Mount(target="/lean-cli-start.sh",
                                          source=str(shell_script_path),
                                          type="bind",
                                          read_only=True))
            kwargs["entrypoint"] = ["bash", "/lean-cli-start.sh"]

        # Format all source paths
        if "mounts" in kwargs:
            for mount in kwargs["mounts"]:
                mount["Source"] = self._format_source_path(mount["Source"])

        if "volumes" in kwargs:
            for key in list(kwargs["volumes"].keys()):
                new_key = self._format_source_path(key)
                kwargs["volumes"][new_key] = kwargs["volumes"].pop(key)

        detach = kwargs.pop("detach", False)
        is_tty = sys.stdout.isatty()

        kwargs["detach"] = True
        kwargs["hostname"] = platform.node()
        kwargs["tty"] = is_tty and not detach
        kwargs["stdin_open"] = is_tty and not detach
        kwargs["stop_signal"] = kwargs.get("stop_signal", "SIGKILL")

        if detach and "remove" not in kwargs:
            kwargs["remove"] = True

        # Make sure host.docker.internal resolves on Linux
        # See https://github.com/QuantConnect/Lean/pull/5092
        if self._platform_manager.is_host_linux():
            if "extra_hosts" not in kwargs:
                kwargs["extra_hosts"] = {}
            kwargs["extra_hosts"]["host.docker.internal"] = "172.17.0.1"

        # Run all containers on a custom bridge network
        # This makes it possible for containers to connect to each other by name
        self.create_network(DOCKER_NETWORK)
        kwargs["network"] = DOCKER_NETWORK

        # Remove existing image with the same name if it exists and is not running
        if "name" in kwargs:
            existing_container = self.get_container_by_name(kwargs["name"])
            if existing_container is not None and existing_container.status != "running":
                existing_container.remove()

        self._logger.debug(f"Running '{image}' with the following configuration:")
        self._logger.debug(kwargs)

        docker_client = self._get_docker_client()
        container = docker_client.containers.run(str(image), None, **kwargs)

        if detach:
            return True

        force_kill_next = False
        killed = False

        def kill_container(force: bool) -> None:
            nonlocal killed
            killed = True
            try:
                if force:
                    container.kill()
                else:
                    container.stop(timeout=60)
                container.remove()
            except APIError:
                pass
            finally:
                self._temp_manager.delete_temporary_directories()
                sys.exit(1)

        # Kill the container on Ctrl+C
        def signal_handler(sig: signal.Signals, frame: types.FrameType) -> None:
            nonlocal force_kill_next
            if not is_tty or force_kill_next or kwargs["stop_signal"] == "SIGKILL":
                force_kill_current = True
            else:
                self._logger.info("Waiting 1 minute for LEAN to exit gracefully, press Ctrl+C again to force stop")
                force_kill_next = True
                force_kill_current = False

            # If we run this code on the current thread, a second Ctrl+C won't be detected on Windows
            kill_thread = threading.Thread(target=kill_container, args=[force_kill_current])
            kill_thread.daemon = True
            kill_thread.start()

        signal.signal(signal.SIGINT, signal_handler)

        # container.logs() is blocking, we run it on a separate thread so the SIGINT handler works properly
        # If we run this code on the current thread, SIGINT won't be triggered on Windows when Ctrl+C is triggered
        def print_logs() -> None:
            chunk_buffer = bytes()
            is_first_time = True
            log_dump = ""

            try:
                while True:
                    container.reload()
                    if container.status != "running":
                        return log_dump

                    if is_first_time:
                        tail = "all"
                        is_first_time = False
                    else:
                        tail = 0

                    # Capture all logs and print it to stdout line by line
                    for chunk in container.logs(stream=True, follow=True, tail=tail):
                        chunk_buffer += chunk

                        if not chunk_buffer.endswith(b"\n"):
                            continue

                        chunk = chunk_buffer.decode("utf-8")
                        chunk_buffer = bytes()

                        if on_output is not None:
                            on_output(chunk)

                        log_dump = log_dump + chunk
                        self._logger.info(chunk.rstrip())

                        if not is_tty:
                            continue

                        if "Press any key to exit..." in chunk or "QuantConnect.Report.Main(): Completed." in chunk:
                            socket = docker_client.api.attach_socket(container.id, params={"stdin": 1, "stream": 1})

                            if hasattr(socket, "_sock"):
                                socket._sock.send(b"\n")
                            else:
                                socket.send(b"\n")

                            socket.close()
            except:
                # This will crash when the container exits, ignore the exception
                pass

        def print_and_format_logs():
            log_dump = print_logs()
            if format_output is not None:
                format_output(log_dump)

        logs_thread = threading.Thread(target=print_and_format_logs)
        logs_thread.daemon = True
        logs_thread.start()

        while logs_thread.is_alive():
            logs_thread.join(0.1)

        if killed:
            try:
                container.remove()
            except APIError:
                pass
            finally:
                sys.exit(1)

        container.wait()

        container.reload()
        success = container.attrs["State"]["ExitCode"] == 0

        container.remove()
        return success

    def build_image(self, root: Path, dockerfile: Path, target: DockerImage) -> None:
        """Builds a Docker image.

        :param root: the path build from
        :param dockerfile: the path to the Dockerfile to build
        :param target: the target name and tag
        """
        # We cannot really use docker_client.images.build() here as it doesn't let us log the progress
        # Building images without showing progress does not provide good developer experience
        # Since the build command is the same on Windows, macOS and Linux we can safely use a system call
        process = subprocess.run(["docker", "build", "-t", str(target), "-f", str(dockerfile), "."], cwd=root)

        if process.returncode != 0:
            raise RuntimeError(
                f"Something went wrong while building '{dockerfile}', see the logs above for more information")

    def image_installed(self, image: DockerImage) -> bool:
        """Returns whether a certain image is installed.

        :param image: the image to check availability for
        :return: True if the image is available locally, False if not
        """
        docker_client = self._get_docker_client()
        return any(str(image) in x.tags for x in docker_client.images.list())

    def get_local_digest(self, image: DockerImage) -> Optional[str]:
        """Returns the digest of a locally installed image.

        :param image: the local image to get the digest of
        :return: the digest of the local image, or None if the digest does not exist
        """
        img = self._get_docker_client().images.get(str(image))

        repo_digests = img.attrs["RepoDigests"]
        if len(repo_digests) == 0:
            return None

        return repo_digests[0].split("@")[1]

    def get_remote_digest(self, image: DockerImage) -> str:
        """Returns the digest of a remote image.

        :param image: the remote image to get the digest of
        :return: the digest of the remote image
        """
        img = self._get_docker_client().images.get_registry_data(str(image))
        return img.attrs["Descriptor"]["digest"]

    def create_network(self, name: str) -> None:
        """Creates a new bridge network, or does nothing if a network with the given name already exists.

        :param name: the name of then network to create
        """
        docker_client = self._get_docker_client()
        if not any(n.name == name for n in docker_client.networks.list()):
            docker_client.networks.create(name, driver="bridge")

    def create_volume(self, name: str) -> None:
        """Creates a new volume, or does nothing if a volume with the given name already exists.

        :param name: the name of the volume to create
        """
        docker_client = self._get_docker_client()
        if not any(v.name == name for v in docker_client.volumes.list()):
            docker_client.volumes.create(name)

    def create_site_packages_volume(self, requirements_file: Path) -> str:
        """Returns the name of the volume to mount to the user's site-packages directory.

        This method automatically returns the best volume for the given requirements.
        It also rotates out older volumes as needed to ensure we don't use too much disk space.

        :param requirements_file: the path to the requirements file that will be pip installed in the container
        :return: the name of the Docker volume to use
        """
        requirements_hash = hashlib.md5(requirements_file.read_text(encoding="utf-8").encode("utf-8")).hexdigest()
        volume_name = f"lean_cli_python_{requirements_hash}"

        docker_client = self._get_docker_client()
        existing_volumes = [v for v in docker_client.volumes.list() if v.name.startswith("lean_cli_python_")]

        if any(v.name == volume_name for v in existing_volumes):
            return volume_name

        volumes_by_age = sorted(existing_volumes, key=lambda v: isoparse(v.attrs["CreatedAt"]))
        for i in range((len(volumes_by_age) - SITE_PACKAGES_VOLUME_LIMIT) + 1):
            volumes_by_age[i].remove()

        docker_client.volumes.create(volume_name)
        return volume_name

    def get_running_containers(self) -> Set[str]:
        """Returns the names of all running containers.

        :return: a set containing the names of all running Docker containers
        """
        containers = self._get_docker_client().containers.list()
        return {c.name.lstrip("/") for c in containers if c.status == "running"}

    def get_container_by_name(self, container_name: str) -> Optional[Container]:
        """Finds a container with a given name.

        :param container_name: the name of the container to find
        :return: the container with the given name, or None if it does not exist
        """
        for container in self._get_docker_client().containers.list(all=True):
            if container.name.lstrip("/") == container_name:
                return container

        return None

    def show_logs(self, container_name: str, follow: bool = False) -> None:
        """Shows the logs of a Docker container in the terminal.

        :param container_name: the name of the container to show the logs of
        :param follow: whether the logs should be streamed in real-time if the container is running (defaults to False)
        """
        if self.get_container_by_name(container_name) is None:
            return

        # We cannot use the Docker Python SDK to get live logs consistently
        # Since the logs command is the same on Windows, macOS and Linux we can safely use a system call
        command = ["docker", "logs"]
        if follow:
            command.append("-f")
        command.append(container_name)

        subprocess.run(command)

    def is_missing_permission(self) -> bool:
        """Returns whether we cannot connect to the Docker client because of a permissions issue.

        A permissions issue usually indicates that the client can only be used with root privileges.

        :return: True if we cannot connect to the Docker client because of a permissions issue, False if that's not
        """
        try:
            docker.from_env()
        except Exception as exception:
            return "Permission denied" in str(exception)
        return False

    def write_to_file(self, docker_container_name: str, docker_file: Path, data: Dict[str, Any]) -> None:
        """Write data to the file in docker.

        Args:
            docker_container_name: The name of the container to write to
            docker_file: The Dockerfile to write to.
            data: The data to write to the Dockerfile.
        """
        docker_container = self.get_container_by_name(docker_container_name)
        if docker_container is None:
            raise ValueError(f"Container {docker_container_name} does not exist")
        if docker_container.status != "running":
            raise ValueError(f"Container {docker_container_name} is not running")

        data = json.dumps(data, cls=DecimalEncoder)
        data = data.replace('"','\\"')
        command = f'docker exec {docker_container_name} bash -c "echo \'{data}\' > {docker_file.as_posix()}"'
        try:
            subprocess.run(command, shell=True, check=True)
        except subprocess.CalledProcessError as exception:
            raise ValueError(f"Failed to write to {docker_file.name}: {exception.output.decode('utf-8')}")
        except Exception as e:
            raise ValueError(f"Failed to write to {docker_file.name}: {e}")

    def read_from_file(self, docker_container_name: str, docker_file: Path, interval=1, timeout=30) -> Dict[str,Any]:
        """Read data from file in docker.

        Args:
            docker_container_name: The name of the container to write to
            docker_file: The Dockerfile to write to.
            interval: The interval to sleep before checking again.
            timeout: The timeout to wait for the file.
        """
        command = f'docker exec {docker_container_name} bash -c "cat {docker_file.as_posix()}"'
        start = time.time()
        success = False
        error_message = None
        container_running = True
        while time.time() - start < timeout:
            try:
                docker_container = self.get_container_by_name(docker_container_name)
                if docker_container is None:
                    error_message = f"Container {docker_container_name} does not exist"
                    container_running = False
                    break
                p = Popen(command, shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE)
                output = p.stdout.read().decode('utf-8')
                if output is not None and output != "":
                    success = True
                    break
            except subprocess.CalledProcessError as exception:
                error_message = f"Failed to read result from docker file {docker_file.name}: {exception.output.decode('utf-8')} {p.stderr.read().decode('utf-8')}"
                time.sleep(interval)
            except Exception as e:
                error_message = f"Failed to read result from docker file {docker_file.name}: {e} {p.stderr.read().decode('utf-8')}"
                time.sleep(interval)

        if success:
            result = json.loads(output)
            success = result["Success"]
            if not success:
                error_message = "Rejected by Lean. Possible arguments error. Please check your logs and try again."
        if not success and not error_message:
            error_message = f"Failed to read result from docker file {docker_file.name} within {timeout} seconds"

        return {
            "error": error_message,
            "success": success,
            "container-running": container_running
        }


    def _get_docker_client(self) -> docker.DockerClient:
        """Creates a DockerClient instance.

        Raises an error if Docker is not running.

        :return: a DockerClient instance which responds to requests
        """
        error = MoreInfoError("Please make sure Docker is installed and running",
                              "https://www.lean.io/docs/v2/lean-cli/key-concepts/troubleshooting#02-Common-Errors")

        try:
            docker_client = docker.from_env()
        except Exception:
            raise error

        try:
            if not docker_client.ping():
                raise error
        except Exception:
            raise error

        return docker_client

    def _format_source_path(self, path: str) -> str:
        """Formats a source path so Docker knows what it refers to.

        This method does two things:
        1. If Docker Toolbox is in use, it converts paths like C:/Path to /c/Path.
        2. If Docker is running in Docker, it converts paths to the corresponding paths on the host system.

        :param path: the original path
        :return: the original path formatted in such a way that Docker can understand it
        """
        # Docker Toolbox modifications
        is_windows = self._platform_manager.is_system_windows()
        is_docker_toolbox = "machine/machines" in os.environ.get("DOCKER_CERT_PATH", "").replace("\\", "/")
        if is_windows and is_docker_toolbox:
            # Backward slashes to forward slashes
            path = path.replace('\\', '/')

            # C:/Path to /c/Path
            path = f"/{path[0].lower()}/{path[3:]}"

        # Docker in Docker modifications
        path_mappings = json.loads(os.environ.get("DOCKER_PATH_MAPPINGS", "{}"))
        for container_path, host_path in path_mappings.items():
            if path.startswith(container_path):
                path = host_path + path[len(container_path):]
                break

        return path



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from lean.components.api.api_client import *
from lean.models.api import QCFullFile, QCMinimalFile


class FileClient:
    """The FileClient class contains methods to interact with files/* API endpoints."""

    def __init__(self, api_client: 'APIClient') -> None:
        """Creates a new FileClient instance.

        :param api_client: the APIClient instance to use when making requests
        """
        self._api = api_client

    def get(self, project_id: int, file_name: str) -> QCMinimalFile:
        """Returns the details of a file.

        :param project_id: the id of the project the file belongs to
        :param file_name: the name of the file to retrieve the details of
        :return: the details of the specified file
        """
        data = self._api.get("files/read", {
            "projectId": project_id,
            "name": file_name
        })

        return QCMinimalFile(**data["files"][0])

    def get_all(self, project_id: int) -> List[QCFullFile]:
        """Returns all files in a project.

        :param project_id: the id of the project to retrieve the files of
        :return: the files in the specified project
        """
        data = self._api.get("files/read", {
            "projectId": project_id
        })

        return [QCFullFile(**file) for file in data["files"]]

    def create(self, project_id: int, file_name: str, content: str) -> QCMinimalFile:
        """Creates a new file.

        :param project_id: the id of the project to create a file for
        :param file_name: the name of the file to create
        :param content: the content of the file to create
        :return: the created file
        """
        data = self._api.post("files/create", {
            "projectId": project_id,
            "name": file_name,
            "content": content
        })

        return QCMinimalFile(**data["files"][0])

    def update(self, project_id: int, file_name: str, content: str) -> QCMinimalFile:
        """Updates an existing file.

        :param project_id: the id of the project to update a file in
        :param file_name: the name of the file to update
        :param content: the new content of the file
        """
        data = self._api.post("files/update", {
            "projectId": project_id,
            "name": file_name,
            "content": content
        })

        return QCMinimalFile(**data["files"][0])

    def delete(self, project_id: int, file_name: str) -> None:
        """Deletes an existing file.

        :param project_id: the id of the project the file belongs to
        :param file_name: the name of the file to delete
        """
        self._api.post("files/delete", {
            "projectId": project_id,
            "name": file_name
        })



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from pathlib import Path

def _capitalize(word: str) -> str:
    """Capitalizes the given word.

    :param word: the word to capitalize
    :return: the word with the first letter capitalized (any other uppercase characters are preserved)
    """
    if word == "":
        return word
    return word[0].upper() + word[1:]


def convert_to_class_name(file_path: Path):
    """Converts the project name into a valid class name by removing all non-alphanumeric characters

    :param file_path: Path to the root project
    :return: returns a valid class name
    """
    return re.sub(f"[^a-zA-Z0-9]", "", "".join(map(_capitalize, file_path.name.split(" "))))


# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from pathlib import Path
from typing import Any, Dict, Optional, List

import json5

from lean.components.cloud.module_manager import ModuleManager
from lean.components.config.cli_config_manager import CLIConfigManager
from lean.components.config.project_config_manager import ProjectConfigManager
from lean.components.config.storage import Storage
from lean.components.util.logger import Logger
from lean.constants import DEFAULT_LEAN_CONFIG_FILE_NAME
from lean.models.errors import MoreInfoError
from lean.models.utils import DebuggingMethod


class LeanConfigManager:
    """The LeanConfigManager class contains utilities to work with files containing LEAN engine configuration."""

    def __init__(self,
                 logger: Logger,
                 cli_config_manager: CLIConfigManager,
                 project_config_manager: ProjectConfigManager,
                 module_manager: ModuleManager,
                 cache_storage: Storage) -> None:
        """Creates a new LeanConfigManager instance.

        :param logger: the logger to log messages with
        :param cli_config_manager: the CLIConfigManager instance to use when retrieving credentials
        :param project_config_manager: the ProjectConfigManager instance to use when retrieving project parameters
        :param module_manager: the ModuleManager to use
        :param cache_storage: the Storage instance to store known Lean config paths in
        """
        self._logger = logger
        self._cli_config_manager = cli_config_manager
        self._project_config_manager = project_config_manager
        self._module_manager = module_manager
        self._cache_storage = cache_storage
        self._default_path = None
        self._lean_config_path = None

    def get_lean_config_path(self) -> Path:
        """Returns the path to the closest Lean config file.

        This recurses upwards in the directory tree looking for a Lean config file.
        This search can be overridden using set_default_lean_config_path().

        Raises an error if no Lean config file can be found.

        :return: the path to the closest Lean config file
        """
        if self._default_path is not None:
            return self._default_path

        if self._lean_config_path is not None:
            return self._lean_config_path

        # Recurse upwards in the directory tree until we find a Lean config file
        current_dir = Path.cwd()
        while True:
            target_file = current_dir / DEFAULT_LEAN_CONFIG_FILE_NAME
            if target_file.exists():
                self._lean_config_path = target_file
                self.store_known_lean_config_path(self._lean_config_path)
                return self._lean_config_path

            # If the parent directory is the same as the current directory we can't go up any more
            if current_dir.parent == current_dir:
                raise MoreInfoError(f"'{DEFAULT_LEAN_CONFIG_FILE_NAME}' not found",
                                    "https://www.lean.io/docs/v2/lean-cli/initialization/configuration#03-Lean-Configuration")

            current_dir = current_dir.parent

    def set_default_lean_config_path(self, path: Path) -> None:
        """Overrides the default search for the path to the Lean config file.

        :param path: the path to the Lean config file to return in future calls to get_lean_config_path()
        """
        self.store_known_lean_config_path(path)
        self._default_path = path

    def get_known_lean_config_paths(self) -> List[Path]:
        """Returns the known Lean config file paths.

        :return: a list of paths to Lean config files that were used in the past
        """
        lean_config_paths = self._cache_storage.get("known-lean-config-paths", [])
        lean_config_paths = [Path(p) for p in lean_config_paths]
        lean_config_paths = [p for p in lean_config_paths if p.is_file()]

        self._cache_storage.set("known-lean-config-paths", [str(p) for p in lean_config_paths])

        return lean_config_paths

    def store_known_lean_config_path(self, path: Path) -> None:
        """Caches a path as being a known Lean configuration file path.

        :param path: the path to the Lean configuration file
        """
        lean_config_paths = set(self._cache_storage.get("known-lean-config-paths", []))
        lean_config_paths.add(str(path))
        self._cache_storage.set("known-lean-config-paths", list(lean_config_paths))

    def get_cli_root_directory(self) -> Path:
        """Returns the path to the directory containing the Lean config file.

        :return: the path to the directory containing the Lean config file
        """
        return self.get_lean_config_path().parent

    def get_data_directory(self) -> Path:
        """Returns the path to the data directory.

        :return: the path to the data directory as it is configured in the Lean config
        """
        config = self.get_lean_config()
        return self.get_cli_root_directory() / config["data-folder"]

    def set_properties(self, updates: Dict[str, Any]) -> None:
        """Sets a properties in the Lean config file.

        If a property does not exist yet it is added automatically.
        Comments in the Lean config file are preserved.

        :param updates: the key -> new value updates to apply to the current config
        """
        config = self.get_lean_config()

        config_path = self.get_lean_config_path()
        config_text = config_path.read_text(encoding="utf-8")

        for key, value in reversed(list(updates.items())):
            json_value = json5.dumps(value)

            # We can only use regex to set the property because converting the config back to JSON drops all comments
            if key in config:
                config_text = re.sub(fr'"{key}":\s*("?[^",]*"?)', f'"{key}": {json_value}', config_text)
            else:
                config_text = config_text.replace("{", f'{{\n  "{key}": {json_value},', 1)

        config_path.write_text(config_text, encoding="utf-8")

    def clean_lean_config(self, config: str) -> str:
        """Removes the properties from a Lean config file which can be set in get_complete_lean_config().

        This removes all the properties which the CLI can configure automatically based on the command that is ran.

        For example, given the following config:
        {
            // Environment docs
            "environment": "backtesting",

            // Key2 docs
            "key2": "value2"
        }

        Calling clean_lean_config(config) would return the following:
        {
            // Key2 docs
            "key2": "value2"
        }

        Because "environment" can be set automatically based on the command that is ran.

        :param config: the configuration to remove the auto-configurable keys from
        :return: the same config as passed in with the config argument, but without the auto-configurable keys
        """
        # The keys that we can set automatically based on the command that is ran
        keys_to_remove = ["environment",
                          "composer-dll-directory",
                          "debugging", "debugging-method",
                          "job-user-id", "api-access-token",
                          "algorithm-type-name", "algorithm-language", "algorithm-location",
                          "parameters", "intrinio-username", "intrinio-password", "ema-fast", "ema-slow"]

        # This function is implemented by doing string manipulation because the config contains comments
        # If we were to parse it as JSON, we would have to remove the comments, which we don't want to do
        sections = re.split(r"\n\s*\n", config)
        for key in keys_to_remove:
            sections = [section for section in sections if f"\"{key}\": " not in section]
        config = "\n\n".join(sections)

        # For some keys we should only remove the key itself, instead of their entire section
        lines = config.split("\n")
        for key in ["ib-host", "ib-port", "ib-tws-dir", "ib-version"]:
            lines = [line for line in lines if f"\"{key}\": " not in line]
        config = "\n".join(lines)

        # Instead of setting the IQFeed host we require the user to set the IQConnect location
        config = config.replace('"iqfeed-host": "127.0.0.1"',
                                '"iqfeed-iqconnect": "C:/Program Files (x86)/DTN/IQFeed/iqconnect.exe"')

        return config

    def get_complete_lean_config(self,
                                 environment: str,
                                 algorithm_file: Path,
                                 debugging_method: Optional[DebuggingMethod]) -> Dict[str, Any]:
        """Returns a complete Lean config object containing all properties needed for the engine to run.

        This retrieves the path of the config, parses the file and adds all properties removed in clean_lean_config().

        :param environment: the environment to set
        :param algorithm_file: the path to the algorithm that will be ran
        :param debugging_method: the debugging method to use, or None to disable debugging
        """
        config = self.get_lean_config()

        config["environment"] = environment
        config["close-automatically"] = True

        config["composer-dll-directory"] = "."

        if debugging_method is not None:
            config["debugging"] = True
            config["debugging-method"] = debugging_method.get_internal_name()
        else:
            config["debugging"] = False
            config["debugging-method"] = "LocalCmdline"

        # The following key -> value pairs are added to the config unless they are already set by the user
        config_defaults = {
            "job-user-id": self._cli_config_manager.user_id.get_value(default="0"),
            "api-access-token": self._cli_config_manager.api_token.get_value(default=""),
            "job-project-id": self._project_config_manager.get_local_id(algorithm_file.parent),

            "ib-host": "127.0.0.1",
            "ib-port": "4002",
            "ib-tws-dir": "/root/Jts",
            "ib-version": "985",

            "iqfeed-host": "host.docker.internal"
        }

        for key, value in config_defaults.items():
            if config.get(key, "") == "":
                config[key] = value

        if algorithm_file.name.endswith(".py"):
            config["algorithm-type-name"] = algorithm_file.name.split(".")[0]
            config["algorithm-language"] = "Python"
            config["algorithm-location"] = f"/LeanCLI/{algorithm_file.name}"
        else:
            algorithm_text = algorithm_file.read_text(encoding="utf-8")
            config["algorithm-type-name"] = re.findall(r"class\s*([^\s:]+)\s*:\s*QCAlgorithm", algorithm_text)[0]
            config["algorithm-language"] = "CSharp"
            config["algorithm-location"] = f"{algorithm_file.parent.name}.dll"

        project_config = self._project_config_manager.get_project_config(algorithm_file.parent)
        config["parameters"] = project_config.get("parameters", {})

        # Add libraries paths to python project
        project_language = project_config.get("algorithm-language", None)
        if project_language == "Python":
            library_references = project_config.get("libraries", [])
            python_paths = config.get("python-additional-paths", [])
            python_paths.extend([(Path("/") / library["path"]).as_posix() for library in library_references])
            if len(python_paths) > 0:
                python_paths.append("/Library")
            config["python-additional-paths"] = python_paths

        # No real limit for the object store by default
        if "storage-limit-mb" not in config:
            config["storage-limit-mb"] = "9999999"
        if "storage-file-count" not in config:
            config["storage-file-count"] = "9999999"

        return config

    def configure_data_purchase_limit(self, lean_config: Dict[str, Any], data_purchase_limit: Optional[int]) -> None:
        """Updates the data purchase limit in the Lean config.

        Logs a warning if the data provider is not configured to download from QuantConnect.

        :param lean_config: the Lean config dict to update
        :param data_purchase_limit: the data purchase limit provided by the user, or None if no such limit was provided
        """
        if data_purchase_limit is None:
            return

        if lean_config.get("data-provider", None) != "QuantConnect.Lean.Engine.DataFeeds.ApiDataProvider":
            self._logger.warn(
                "--data-purchase-limit is ignored because the data provider is not set to download from the QuantConnect API, use --download-data to set that up")
            return

        lean_config["data-purchase-limit"] = data_purchase_limit

    def get_lean_config(self) -> Dict[str, Any]:
        """Reads the Lean config into a dict.

        :return: a dict containing the contents of the Lean config file
        """
        return json5.loads(self.get_lean_config_path().read_text(encoding="utf-8"))



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from pathlib import Path
from typing import Any, Dict, Optional

import click
from docker.types import Mount

from lean.click import LeanCommand, PathParameter
from lean.constants import DEFAULT_ENGINE_IMAGE, PROJECT_CONFIG_FILE_NAME
from lean.container import container
from lean.models.errors import MoreInfoError
from lean.components.util.live_utils import get_state_json


def _find_project_directory(backtest_file: Path) -> Optional[Path]:
    """Returns the project directory, or None if backtest_file is not stored in a project directory.

    :param backtest_file: the path to the JSON file containing the backtest results
    :return: the path to the project directory, or None if backtest_file is stored outside a project directory
    """
    current_directory = backtest_file.parent

    # Loop until we find the root directory
    while current_directory != current_directory.parent:
        if (current_directory / PROJECT_CONFIG_FILE_NAME).is_file():
            return current_directory

        current_directory = current_directory.parent

    return None


@click.command(cls=LeanCommand, requires_lean_config=True, requires_docker=True)
@click.option("--backtest-results",
              type=PathParameter(exists=True, file_okay=True, dir_okay=False),
              help="Path to the JSON file containing the backtest results")
@click.option("--live-results",
              type=PathParameter(exists=True, file_okay=True, dir_okay=False),
              help="Path to the JSON file containing the live trading results")
@click.option("--report-destination",
              type=PathParameter(exists=False, file_okay=True, dir_okay=False),
              default=lambda: Path.cwd() / "report.html",
              help="Path where the generated report is stored as HTML (defaults to ./report.html)")
@click.option("--detach", "-d",
              is_flag=True,
              default=False,
              help="Run the report creator in a detached Docker container and return immediately")
@click.option("--strategy-name",
              type=str,
              help="Name of the strategy, will appear at the top-right corner of each page")
@click.option("--strategy-version",
              type=str,
              help="Version number of the strategy, will appear next to the project name")
@click.option("--strategy-description",
              type=str,
              help="Description of the strategy, will appear under the 'Strategy Description' section")
@click.option("--overwrite",
              is_flag=True,
              default=False,
              help="Overwrite --report-destination if it already contains a file")
@click.option("--image",
              type=str,
              help=f"The LEAN engine image to use (defaults to {DEFAULT_ENGINE_IMAGE})")
@click.option("--update",
              is_flag=True,
              default=False,
              help="Pull the LEAN engine image before running the report creator")
def report(backtest_results: Optional[Path],
           live_results: Optional[Path],
           report_destination: Path,
           detach: bool,
           strategy_name: Optional[str],
           strategy_version: Optional[str],
           strategy_description: Optional[str],
           overwrite: bool,
           image: Optional[str],
           update: bool) -> None:
    """Generate a report of a backtest.

    This runs the LEAN Report Creator in Docker to generate a polished, professional-grade report of a backtest.

    If --backtest-results is not given, a report is generated for the most recent local backtest.

    The name, description, and version are optional and will be blank if not given.

    If the given backtest data source file is stored in a project directory (or one of its subdirectories, like the
    default <project>/backtests/<timestamp>), the default name is the name of the project directory and the default
    description is the description stored in the project's config.json file.

    By default the official LEAN engine image is used.
    You can override this using the --image option.
    Alternatively you can set the default engine image for all commands using `lean config set engine-image <image>`.
    """
    if report_destination.exists() and not overwrite:
        raise RuntimeError(f"{report_destination} already exists, use --overwrite to overwrite it")

    if backtest_results is None:
        backtest_results = get_state_json("backtests")
        if not backtest_results:
            raise MoreInfoError(
            "Could not find a recent backtest result file, please use the --backtest-results option",
            "https://www.lean.io/docs/v2/lean-cli/reports#02-Generate-Reports"
            )

    logger = container.logger()

    if live_results is None:
        logger.info(f"Generating a report from '{backtest_results}'")
    else:
        logger.info(f"Generating a report from '{backtest_results}' and '{live_results}'")

    project_directory = _find_project_directory(backtest_results)

    if project_directory is not None:
        if strategy_name is None:
            strategy_name = project_directory.name

        if strategy_description is None:
            project_config_manager = container.project_config_manager()
            project_config = project_config_manager.get_project_config(project_directory)
            strategy_description = project_config.get("description", "")

    # The configuration given to the report creator
    # See https://github.com/QuantConnect/Lean/blob/master/Report/config.example.json
    report_config = {
        "data-folder": "/Lean/Data",
        "strategy-name": strategy_name or "",
        "strategy-version": strategy_version or "",
        "strategy-description": strategy_description or "",
        "live-data-source-file": "live-data-source-file.json" if live_results is not None else "",
        "backtest-data-source-file": "backtest-data-source-file.json",
        "report-destination": "/tmp/report.html",

        "environment": "report",

        "log-handler": "QuantConnect.Logging.CompositeLogHandler",
        "messaging-handler": "QuantConnect.Messaging.Messaging",
        "job-queue-handler": "QuantConnect.Queues.JobQueue",
        "api-handler": "QuantConnect.Api.Api",
        "map-file-provider": "QuantConnect.Data.Auxiliary.LocalDiskMapFileProvider",
        "factor-file-provider": "QuantConnect.Data.Auxiliary.LocalDiskFactorFileProvider",
        "data-provider": "QuantConnect.Lean.Engine.DataFeeds.DefaultDataProvider",
        "alpha-handler": "QuantConnect.Lean.Engine.Alphas.DefaultAlphaHandler",
        "data-channel-provider": "DataChannelProvider",

        "environments": {
            "report": {
                "live-mode": False,

                "setup-handler": "QuantConnect.Lean.Engine.Setup.ConsoleSetupHandler",
                "result-handler": "QuantConnect.Lean.Engine.Results.BacktestingResultHandler",
                "data-feed-handler": "QuantConnect.Lean.Engine.DataFeeds.FileSystemDataFeed",
                "real-time-handler": "QuantConnect.Lean.Engine.RealTime.BacktestingRealTimeHandler",
                "history-provider": "QuantConnect.Lean.Engine.HistoricalData.SubscriptionDataReaderHistoryProvider",
                "transaction-handler": "QuantConnect.Lean.Engine.TransactionHandlers.BacktestingTransactionHandler"
            }
        }
    }

    config_path = container.temp_manager().create_temporary_directory() / "config.json"
    with config_path.open("w+", encoding="utf-8") as file:
        json.dump(report_config, file)

    backtest_id = container.output_config_manager().get_backtest_id(backtest_results.parent)

    lean_config_manager = container.lean_config_manager()
    data_dir = lean_config_manager.get_data_directory()

    report_destination.parent.mkdir(parents=True, exist_ok=True)

    run_options: Dict[str, Any] = {
        "detach": detach,
        "name": f"lean_cli_report_{backtest_id}",
        "working_dir": "/Lean/Report/bin/Debug",
        "commands": ["dotnet QuantConnect.Report.dll", f'cp /tmp/report.html "/Output/{report_destination.name}"'],
        "mounts": [
            Mount(target="/Lean/Report/bin/Debug/config.json",
                  source=str(config_path),
                  type="bind",
                  read_only=True),
            Mount(target="/Lean/Report/bin/Debug/backtest-data-source-file.json",
                  source=str(backtest_results),
                  type="bind",
                  read_only=True)
        ],
        "volumes": {
            str(data_dir): {
                "bind": "/Lean/Data",
                "mode": "rw"
            },
            str(report_destination.parent): {
                "bind": "/Output",
                "mode": "rw"
            }
        }
    }

    if live_results is not None:
        run_options["mounts"].append(Mount(target="/Lean/Report/bin/Debug/live-data-source-file.json",
                                           source=str(live_results),
                                           type="bind",
                                           read_only=True))

    cli_config_manager = container.cli_config_manager()
    engine_image_override = image

    if engine_image_override is None and project_directory is not None:
        project_config_manager = container.project_config_manager()
        project_config = project_config_manager.get_project_config(project_directory)
        engine_image_override = project_config.get("engine-image", None)

    engine_image = cli_config_manager.get_engine_image(engine_image_override)

    container.update_manager().pull_docker_image_if_necessary(engine_image, update)

    success = container.docker_manager().run_image(engine_image, **run_options)
    if not success:
        raise RuntimeError(
            "Something went wrong while running the LEAN Report Creator, see the logs above for more information")

    if detach:
        temp_manager = container.temp_manager()
        temp_manager.delete_temporary_directories_when_done = False

        logger.info(f"Successfully started the report creator in the '{run_options['name']}' container")
        logger.info(f"The report will be generated to '{report_destination}'")
        logger.info("You can use Docker's own commands to manage the detached container")
        return

    logger.info(f"Successfully generated report to '{report_destination}'")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import requests


class RequestFailedError(Exception):
    """A RequestFailedError indicates that an HTTP request has failed."""

    def __init__(self, response: requests.Response, message: Optional[str] = None) -> None:
        """Creates a new RequestFailedError instance.

        :param response: the data of the failed response
        :param message: a display-friendly error message, defaults to a message based on the response
        """
        if message is None:
            request = response.request
            message = f"{request.method} request to {request.url} failed with status code {response.status_code}"

        super().__init__(message)

        self.response = response


class MoreInfoError(Exception):
    """An error which consists of a message and a link to documentation with more information."""

    def __init__(self, message: str, link: str) -> None:
        """Creates a new MoreInfoError instance.

        :param message: the error message
        :param link: the link to documentation containing more information
        """
        super().__init__(message)

        self.link = link


class AuthenticationError(MoreInfoError):
    """An error indicating that a request has failed because the used credentials were invalid."""

    def __init__(self) -> None:
        """Creates a new AuthenticationError instance."""
        super().__init__("Invalid credentials, please log in using `lean login`",
                         "https://www.lean.io/docs/v2/lean-cli/initialization/authenticating-accounts#02-Log-In")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from lean.commands.config.get import get
from lean.commands.config.list import list
from lean.commands.config.set import set
from lean.commands.config.unset import unset


@click.group()
def config() -> None:
    """Configure Lean CLI options."""
    # This method is intentionally empty
    # It is used as the command group for all `lean config <command>` commands
    pass


config.add_command(get)
config.add_command(set)
config.add_command(unset)
config.add_command(list)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from math import floor
from typing import List, Optional

from lean.components.api.api_client import *
from lean.models.api import QCFullLiveAlgorithm, QCLiveAlgorithmStatus, QCMinimalLiveAlgorithm, QCNotificationMethod, QCRestResponse


class LiveClient:
    """The LiveClient class contains methods to interact with live/* API endpoints."""

    def __init__(self, api_client: 'APIClient') -> None:
        """Creates a new LiveClient instance.

        :param api_client: the APIClient instance to use when making requests
        """
        self._api = api_client

    def get_all(self,
                status: Optional[QCLiveAlgorithmStatus] = None,
                # Values less than 86400 cause errors on Windows: https://bugs.python.org/issue37527
                start: datetime = datetime.fromtimestamp(86400),
                end: datetime = datetime.now()) -> List[QCFullLiveAlgorithm]:
        """Retrieves all live algorithms.

        :param status: the status to filter by or None if no status filter should be applied
        :param start: the earliest launch time the returned algorithms should have
        :param end: the latest launch time the returned algorithms should have
        :return: a list of live algorithms which match the given filters
        """
        parameters = {
            "start": floor(start.timestamp()),
            "end": floor(end.timestamp())
        }

        if status is not None:
            parameters["status"] = status.value

        data = self._api.get("live/read", parameters)
        return [QCFullLiveAlgorithm(**algorithm) for algorithm in data["live"]]

    def start(self,
              project_id: int,
              compile_id: str,
              node_id: str,
              brokerage_settings: Dict[str, Any],
              price_data_handler: str,
              automatic_redeploy: bool,
              version_id: int,
              notify_order_events: bool,
              notify_insights: bool,
              notify_methods: List[QCNotificationMethod],
              live_cash_balance: Optional[List[Dict[str, float]]] = None,
              live_holdings: Optional[List[Dict[str, float]]] = None) -> QCMinimalLiveAlgorithm:
        """Starts live trading for a project.

        :param project_id: the id of the project to start live trading for
        :param compile_id: the id of the compile to use for live trading
        :param node_id: the id of the node to start live trading on
        :param brokerage_settings: the brokerage settings to use
        :param price_data_handler: the data feed to use
        :param automatic_redeploy: whether automatic redeploys are enabled
        :param version_id: the id of the LEAN version to use
        :param notify_order_events: whether notifications should be sent on order events
        :param notify_insights: whether notifications should be sent on insights
        :param notify_methods: the places to send notifications to
        :param live_cash_balance: the list of initial cash balance
        :param live_holdings: the list of initial portfolio holdings
        :return: the created live algorithm
        """

        if live_cash_balance:
            brokerage_settings["cash"] = live_cash_balance
        if live_holdings:
            brokerage_settings["holdings"] = live_holdings

        parameters = {
            "projectId": project_id,
            "compileId": compile_id,
            "nodeId": node_id,
            "brokerage": brokerage_settings,
            "dataHandler": price_data_handler,
            "automaticRedeploy": automatic_redeploy,
            "versionId": version_id
        }

        if notify_order_events or notify_insights:
            events = []
            if notify_order_events:
                events.append("orderEvent")
            if notify_insights:
                events.append("insight")

            parameters["notification"] = {
                "events": events,
                "targets": [{x: y for x, y in method.dict().items() if y} for method in notify_methods]
            }

        data = self._api.post("live/create", parameters)
        return QCMinimalLiveAlgorithm(**data)

    def stop(self, project_id: int) -> QCRestResponse:
        """Stops live trading for a certain project without liquidated existing positions.

        :param project_id: the id of the project to stop live trading for
        """
        data = self._api.post("live/update/stop", {
            "projectId": project_id
        })
        return QCRestResponse(**data)

    def liquidate_and_stop(self, project_id: int) -> QCRestResponse:
        """Stops live trading and liquidates existing positions for a certain project.

        :param project_id: the id of the project to stop live trading and liquidate existing positions for
        """
        data = self._api.post("live/update/liquidate", {
            "projectId": project_id
        })
        return QCRestResponse(**data)


# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
import uuid
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional, List

from docker.types import Mount
from pkg_resources import Requirement

from lean.components.cloud.module_manager import ModuleManager
from lean.components.config.lean_config_manager import LeanConfigManager
from lean.components.config.output_config_manager import OutputConfigManager
from lean.components.config.project_config_manager import ProjectConfigManager
from lean.components.docker.docker_manager import DockerManager
from lean.components.util.logger import Logger
from lean.components.util.project_manager import ProjectManager
from lean.components.util.temp_manager import TempManager
from lean.components.util.xml_manager import XMLManager
from lean.constants import MODULES_DIRECTORY, TERMINAL_LINK_PRODUCT_ID, LEAN_ROOT_PATH
from lean.constants import DOCKER_PYTHON_SITE_PACKAGES_PATH
from lean.models.docker import DockerImage
from lean.models.utils import DebuggingMethod


class LeanRunner:
    """The LeanRunner class contains the code that runs the LEAN engine locally."""

    def __init__(self,
                 logger: Logger,
                 project_config_manager: ProjectConfigManager,
                 lean_config_manager: LeanConfigManager,
                 output_config_manager: OutputConfigManager,
                 docker_manager: DockerManager,
                 module_manager: ModuleManager,
                 project_manager: ProjectManager,
                 temp_manager: TempManager,
                 xml_manager: XMLManager) -> None:
        """Creates a new LeanRunner instance.

        :param logger: the logger that is used to print messages
        :param project_config_manager: the ProjectConfigManager instance to retrieve project configuration from
        :param lean_config_manager: the LeanConfigManager instance to retrieve Lean configuration from
        :param output_config_manager: the OutputConfigManager instance to retrieve backtest/live configuration from
        :param docker_manager: the DockerManager instance which is used to interact with Docker
        :param module_manager: the ModuleManager instance to retrieve the installed modules from
        :param project_manager: the ProjectManager instance to use for copying source code to output directories
        :param temp_manager: the TempManager instance to use for creating temporary directories
        :param xml_manager: the XMLManager instance to use for reading/writing XML files
        """
        self._logger = logger
        self._project_config_manager = project_config_manager
        self._lean_config_manager = lean_config_manager
        self._output_config_manager = output_config_manager
        self._docker_manager = docker_manager
        self._module_manager = module_manager
        self._project_manager = project_manager
        self._temp_manager = temp_manager
        self._xml_manager = xml_manager

    def run_lean(self,
                 lean_config: Dict[str, Any],
                 environment: str,
                 algorithm_file: Path,
                 output_dir: Path,
                 image: DockerImage,
                 debugging_method: Optional[DebuggingMethod],
                 release: bool,
                 detach: bool) -> None:
        """Runs the LEAN engine locally in Docker.

        Raises an error if something goes wrong.

        :param lean_config: the LEAN configuration to use
        :param environment: the environment to run the algorithm in
        :param algorithm_file: the path to the file containing the algorithm
        :param output_dir: the directory to save output data to
        :param image: the LEAN engine image to use
        :param debugging_method: the debugging method if debugging needs to be enabled, None if not
        :param release: whether C# projects should be compiled in release configuration instead of debug
        :param detach: whether LEAN should run in a detached container
        """
        project_dir = algorithm_file.parent

        # The dict containing all options passed to `docker run`
        # See all available options at https://docker-py.readthedocs.io/en/stable/containers.html
        run_options = self.get_basic_docker_config(lean_config,
                                                   algorithm_file,
                                                   output_dir,
                                                   debugging_method,
                                                   release,
                                                   detach)

        # Set up PTVSD debugging
        if debugging_method == DebuggingMethod.PTVSD:
            run_options["ports"]["5678"] = "5678"

        # Set up VSDBG debugging
        if debugging_method == DebuggingMethod.VSDBG:
            run_options["name"] = "lean_cli_vsdbg"

            # lean_cli_vsdbg is not unique, so we don't store the container name in the output directory's config
            output_config = self._output_config_manager.get_output_config(output_dir)
            output_config.delete("container")

        # Set up Rider debugging
        if debugging_method == DebuggingMethod.Rider:
            run_options["ports"]["22"] = "2222"
            run_options["commands"].append("/usr/sbin/enable_insecure_key")
            run_options["commands"].append("chmod 600 /etc/insecure_key")
            run_options["commands"].append('echo "HostKey /etc/insecure_key" >> /etc/ssh/sshd_config')
            run_options["commands"].append("/usr/sbin/sshd")

            self._docker_manager.create_volume("lean_cli_rider")
            run_options["volumes"]["lean_cli_rider"] = {
                "bind": "/root/.local/share/JetBrains",
                "mode": "rw"
            }

        run_options["commands"].append("exec dotnet QuantConnect.Lean.Launcher.dll")

        # Copy the project's code to the output directory
        self._project_manager.copy_code(algorithm_file.parent, output_dir / "code")

        # Run the engine and log the result
        # Run as a subprocess to capture the output before logging it

        # Format error messages for cleaner output logs
        run_options["format_output"] = self.format_error_before_logging

        success = self._docker_manager.run_image(image, **run_options)

        cli_root_dir = self._lean_config_manager.get_cli_root_directory()
        relative_project_dir = project_dir.relative_to(cli_root_dir)
        relative_output_dir = output_dir.relative_to(cli_root_dir)

        if detach:
            self._temp_manager.delete_temporary_directories_when_done = False

            self._logger.info(
                f"Successfully started '{relative_project_dir}' in the '{environment}' environment in the '{run_options['name']}' container")
            self._logger.info(f"The output will be stored in '{relative_output_dir}'")
            self._logger.info("You can use Docker's own commands to manage the detached container")
        elif success:
            self._logger.info(
                f"Successfully ran '{relative_project_dir}' in the '{environment}' environment and stored the output in '{relative_output_dir}'")
        else:
            raise RuntimeError(
                f"Something went wrong while running '{relative_project_dir}' in the '{environment}' environment, the output is stored in '{relative_output_dir}'")

    def get_basic_docker_config(self,
                                lean_config: Dict[str, Any],
                                algorithm_file: Path,
                                output_dir: Path,
                                debugging_method: Optional[DebuggingMethod],
                                release: bool,
                                detach: bool) -> Dict[str, Any]:
        """Creates a basic Docker config to run the engine with.

        This method constructs the parts of the Docker config that is the same for both the engine and the optimizer.

        :param lean_config: the LEAN configuration to use
        :param algorithm_file: the path to the file containing the algorithm
        :param output_dir: the directory to save output data to
        :param debugging_method: the debugging method if debugging needs to be enabled, None if not
        :param release: whether C# projects should be compiled in release configuration instead of debug
        :param detach: whether LEAN should run in a detached container
        :return: the Docker configuration containing basic configuration to run Lean
        """
        project_dir = algorithm_file.parent
        project_config = self._project_config_manager.get_project_config(project_dir)
        docker_project_config = project_config.get("docker", {})

        # Install the required modules when they're needed
        if lean_config.get("data-provider", None) == "QuantConnect.Lean.Engine.DataFeeds.DownloaderDataProvider" \
            and lean_config.get("data-downloader", None) == "TerminalLinkDataDownloader":
            self._module_manager.install_module(TERMINAL_LINK_PRODUCT_ID, lean_config["job-organization-id"])

        # Force the use of the LocalDisk map/factor providers if no recent zip present and not using ApiDataProvider
        data_dir = self._lean_config_manager.get_data_directory()
        if lean_config.get("data-provider", None) != "QuantConnect.Lean.Engine.DataFeeds.ApiDataProvider":
            self._force_disk_provider_if_necessary(lean_config,
                                                   "map-file-provider",
                                                   "QuantConnect.Data.Auxiliary.LocalZipMapFileProvider",
                                                   "QuantConnect.Data.Auxiliary.LocalDiskMapFileProvider",
                                                   data_dir / "equity" / "usa" / "map_files")
            self._force_disk_provider_if_necessary(lean_config,
                                                   "factor-file-provider",
                                                   "QuantConnect.Data.Auxiliary.LocalZipFactorFileProvider",
                                                   "QuantConnect.Data.Auxiliary.LocalDiskFactorFileProvider",
                                                   data_dir / "equity" / "usa" / "factor_files")

        # Create the output directory if it doesn't exist yet
        if not output_dir.exists():
            output_dir.mkdir(parents=True)

        # Create the storage directory if it doesn't exist yet
        storage_dir = project_dir / "storage"
        if not storage_dir.exists():
            storage_dir.mkdir(parents=True)

        lean_config["debug-mode"] = self._logger.debug_logging_enabled
        lean_config["data-folder"] = "/Lean/Data"
        lean_config["results-destination-folder"] = "/Results"
        lean_config["object-store-root"] = "/Storage"

        # The dict containing all options passed to `docker run`
        # See all available options at https://docker-py.readthedocs.io/en/stable/containers.html
        run_options: Dict[str, Any] = {
            "detach": detach,
            "commands": [],
            "environment": docker_project_config.get("environment", {}),
            "stop_signal": "SIGINT" if debugging_method is None else "SIGKILL",
            "mounts": [],
            "volumes": {},
            "ports": docker_project_config.get("ports", {})
        }

        # Mount the project directory
        run_options["volumes"][str(project_dir)] = {
            "bind": "/LeanCLI",
            "mode": "rw"
        }

        # Check if the user has library projects and mount the Library directory
        library_dir = self._lean_config_manager.get_cli_root_directory() / "Library"
        if library_dir.is_dir():
            # Mount the library projects
            run_options["volumes"][str(library_dir)] = {
                "bind": "/Library",
                "mode": "rw"
            }

        # Mount the data directory
        run_options["volumes"][str(data_dir)] = {
            "bind": "/Lean/Data",
            "mode": "rw"
        }

        # Mount the output directory
        run_options["volumes"][str(output_dir)] = {
            "bind": "/Results",
            "mode": "rw"
        }

        # Mount the local object store directory
        run_options["volumes"][str(storage_dir)] = {
            "bind": "/Storage",
            "mode": "rw"
        }

        # Mount all local files referenced in the Lean config
        cli_root_dir = self._lean_config_manager.get_cli_root_directory()
        for key in ["transaction-log", "terminal-link-symbol-map-file"]:
            if key not in lean_config or lean_config[key] == "":
                continue

            local_path = cli_root_dir / lean_config[key]
            if not local_path.exists():
                local_path.parent.mkdir(parents=True, exist_ok=True)
                local_path.touch()

            run_options["mounts"].append(Mount(target=f"/Files/{key}",
                                               source=str(local_path),
                                               type="bind",
                                               read_only=False))

            lean_config[key] = f"/Files/{key}"

        # Update all hosts that need to point to the host's localhost to host.docker.internal so they resolve properly
        for key in ["terminal-link-server-host"]:
            if key not in lean_config:
                continue

            if lean_config[key] == "localhost" or lean_config[key] == "127.0.0.1":
                lean_config[key] = "host.docker.internal"

        set_up_common_csharp_options_called = False

        # Set up modules
        installed_packages = self._module_manager.get_installed_packages()
        if len(installed_packages) > 0:
            self._logger.debug(f"LeanRunner.run_lean(): installed packages {len(installed_packages)}")
            self.set_up_common_csharp_options(run_options)
            set_up_common_csharp_options_called = True

            # Mount the modules directory
            run_options["volumes"][MODULES_DIRECTORY] = {
                "bind": "/Modules",
                "mode": "ro"
            }

            # Add the modules directory as a NuGet source root
            run_options["commands"].append("dotnet nuget add source /Modules")

            # Create a C# project used to resolve the dependencies of the modules
            run_options["commands"].append("mkdir /ModulesProject")
            run_options["commands"].append("dotnet new sln -o /ModulesProject")
            run_options["commands"].append("dotnet new classlib -o /ModulesProject -f net6.0 --no-restore")
            run_options["commands"].append("rm /ModulesProject/Class1.cs")

            # Add all modules to the project, automatically resolving all dependencies
            for package in installed_packages:
                self._logger.debug(f"LeanRunner.run_lean(): Adding module {package} to the project")
                run_options["commands"].append(f"rm -rf /root/.nuget/packages/{package.name.lower()}")
                run_options["commands"].append(
                    f"dotnet add /ModulesProject package {package.name} --version {package.version}")

            # Copy all module files to /Lean/Launcher/bin/Debug, but don't overwrite anything that already exists
            run_options["commands"].append(
                "python /copy_csharp_dependencies.py /Compile/obj/ModulesProject/project.assets.json")

        # Set up language-specific run options
        self.setup_language_specific_run_options(run_options, project_dir, algorithm_file,
                                            set_up_common_csharp_options_called, release)

        # Save the final Lean config to a temporary file so we can mount it into the container
        config_path = self._temp_manager.create_temporary_directory() / "config.json"
        with config_path.open("w+", encoding="utf-8") as file:
            file.write(json.dumps(lean_config, indent=4))

        # Mount the Lean config
        run_options["mounts"].append(Mount(target=f"{LEAN_ROOT_PATH}/config.json",
                                           source=str(config_path),
                                           type="bind",
                                           read_only=True))

        # Assign the container a name and store it in the output directory's configuration
        run_options["name"] = f"lean_cli_{str(uuid.uuid4()).replace('-', '')}"
        output_config = self._output_config_manager.get_output_config(output_dir)
        output_config.set("container", run_options["name"])
        if "backtest-name" in lean_config:
            output_config.set("backtest-name", lean_config["backtest-name"])

        if "environment" in lean_config and "environments" in lean_config:
            environment = lean_config["environments"][lean_config["environment"]]
            if "live-mode-brokerage" in environment:
                output_config.set("brokerage", environment["live-mode-brokerage"].split(".")[-1])

        return run_options

    def set_up_python_options(self, project_dir: Path, run_options: Dict[str, Any]) -> None:
        """Sets up Docker run options specific to Python projects.

        :param project_dir: the path to the project directory
        :param run_options: the dictionary to append run options to
        """

        # Compile python files
        source_files = self._project_manager.get_source_files(project_dir)
        source_files = [file.relative_to(
            project_dir).as_posix() for file in source_files]
        source_files = [f'"/LeanCLI/{file}"' for file in source_files]

        # Only need to compile files in backtest/live (where the files were mounted in "/LeanCLI") but not research
        run_options["commands"].append(
            f"""if [ -d '/LeanCLI' ];
            then
                python -m compileall {' '.join(source_files)};
            else
                echo '/LeanCLI is not mounted, skipping compilation...';
            fi""")

        # Combine the requirements from all library projects and the current project
        library_dir = self._lean_config_manager.get_cli_root_directory() / "Library"
        requirements_files = list(library_dir.rglob("requirements.txt")) + [project_dir / "requirements.txt"]
        requirements_files = [file for file in requirements_files if file.is_file()]
        requirements = self._concat_python_requirements(requirements_files)

        # Check if we have any dependencies to install, so we don't mount volumes needlessly
        if requirements == "":
            return

        # Create a requirements.txt file for the combined requirements
        requirements_txt = self._temp_manager.create_temporary_directory() / "requirements.txt"
        with requirements_txt.open("w+", encoding="utf-8") as file:
            file.write(requirements)

        # Mount the requirements.txt file
        run_options["mounts"].append(Mount(target="/requirements.txt",
                                           source=str(requirements_txt),
                                           type="bind",
                                           read_only=True))

        # Mount a volume to pip's cache directory so we only download packages once
        self._docker_manager.create_volume("lean_cli_pip")
        run_options["volumes"]["lean_cli_pip"] = {
            "bind": "/root/.cache/pip",
            "mode": "rw"
        }

        # Mount a volume to the user packages directory so we don't install packages every time
        site_packages_volume = self._docker_manager.create_site_packages_volume(requirements_txt)
        run_options["volumes"][site_packages_volume] = {
            "bind": f"{DOCKER_PYTHON_SITE_PACKAGES_PATH}",
            "mode": "rw"
        }

        # Update PATH in the Docker container to prevent pip install warnings about its executables not being on PATH
        run_options["commands"].append('export PATH="$PATH:/root/.local/bin"')

        # Install custom libraries to the cached user packages directory
        # We only need to do this if it hasn't already been done before for this site packages volume
        # To keep track of this we create a special file in the site packages directory after installation
        # If this file already exists we can skip pip install completely
        marker_file = f"{DOCKER_PYTHON_SITE_PACKAGES_PATH}/pip-install-done"
        run_options["commands"].extend([
            f"! test -f {marker_file} && pip install --user --progress-bar off -r /requirements.txt",
            f"touch {marker_file}"
        ])

    def _concat_python_requirements(self, requirements_files: List[Path]) -> str:
        """Combines the requirements from multiple requirements.txt files.

        Ensures there are no duplicate requirements and that all output lines are valid.
        Requirements are sorted alphabetically to ensure consistent output.

        :param requirements_files: the paths to the requirements.txt files
        :return: the normalized requirements from all given requirements.txt files
        """
        requirements = []
        for file in requirements_files:
            for line in file.read_text(encoding="utf-8").splitlines():
                try:
                    requirements.append(Requirement.parse(line))
                except ValueError:
                    pass

        requirements = [str(requirement) for requirement in requirements]
        requirements = sorted(set(requirements))
        return "\n".join(requirements)

    def set_up_csharp_options(self, project_dir: Path, run_options: Dict[str, Any], release: bool) -> None:
        """Sets up Docker run options specific to C# projects.

        :param project_dir: the path to the project directory
        :param run_options: the dictionary to append run options to
        :param release: whether C# projects should be compiled in release configuration instead of debug
        """
        compile_root = self._get_csharp_compile_root(project_dir)

        # Ensure all .csproj files refer to the version of LEAN in the Docker container
        csproj_temp_dir = self._temp_manager.create_temporary_directory()
        for path in compile_root.rglob("*.csproj"):
            self._ensure_csproj_uses_correct_lean(compile_root, path, csproj_temp_dir, run_options)

        # Set up the MSBuild properties
        msbuild_properties = {
            "Configuration": "Release" if release else "Debug",
            "Platform": "AnyCPU",
            "TargetFramework": "net6.0",
            "OutputPath": "/Compile/bin",
            "GenerateAssemblyInfo": "false",
            "GenerateTargetFrameworkAttribute": "false",
            "AppendTargetFrameworkToOutputPath": "false",
            "AutoGenerateBindingRedirects": "true",
            "GenerateBindingRedirectsOutputType": "true",
            "AutomaticallyUseReferenceAssemblyPackages": "false",
            "CopyLocalLockFileAssemblies": "true",
            "PathMap": f"/LeanCLI={str(compile_root)}",
            "NoWarn": ["MSB3277"]
        }

        # Find the .csproj file to compile
        project_file = next(project_dir.glob("*.csproj"))

        # Inherit NoWarn from the user's .csproj
        csproj = self._xml_manager.parse(project_file.read_text(encoding="utf-8"))
        existing_no_warn = csproj.find(".//NoWarn")
        if existing_no_warn is not None:
            codes = [c for c in re.split(r"[^a-zA-Z0-9]", existing_no_warn.text) if c != ""]
            msbuild_properties["NoWarn"] += codes

        # Turn the NoWarn property into a string
        # %3B is the encoded version of ";", because a raw ";" is seen as a separator between properties
        msbuild_properties["NoWarn"] = "%3B".join(msbuild_properties["NoWarn"])

        # Build the project before running LEAN
        relative_project_file = str(project_file.relative_to(compile_root)).replace("\\", "/")
        msbuild_properties = ";".join(f"{key}={value}" for key, value in msbuild_properties.items())
        run_options["commands"].append(f'dotnet build "/LeanCLI/{relative_project_file}" "-p:{msbuild_properties}"')

        # Copy over the algorithm DLL
        # Copy over the project reference DLLs'
        # Copy over all output DLLs that don't already exist in /Lean/Launcher/bin/Debug
        run_options["commands"].append(f"cp -R -n /Compile/bin/. {LEAN_ROOT_PATH}/")

        # Copy over all library DLLs that don't already exist in /Lean/Launcher/bin/Debug
        # CopyLocalLockFileAssemblies does not copy the OS-specific DLLs to the output directory
        # We therefore use a custom Python script that does take the OS into account when deciding what to copy
        run_options["commands"].append(
            f'python /copy_csharp_dependencies.py "/Compile/obj/{project_file.stem}/project.assets.json"')

    def set_up_common_csharp_options(self, run_options: Dict[str, Any]) -> None:
        """Sets up common Docker run options that is needed for all C# work.

        This method is only called if the user has installed modules and/or if the project to run is written in C#.

        :param run_options: the dictionary to append run options to
        """
        # Mount a volume to NuGet's cache directory so we only download packages once
        self._docker_manager.create_volume("lean_cli_nuget")
        run_options["volumes"]["lean_cli_nuget"] = {
            "bind": "/root/.nuget/packages",
            "mode": "rw"
        }

        # Reduce the dotnet output
        run_options["environment"]["DOTNET_NOLOGO"] = "true"
        run_options["environment"]["DOTNET_CLI_TELEMETRY_OPTOUT"] = "true"

        temp_files_directory = self._temp_manager.create_temporary_directory()

        # Create a Directory.Build.props file to ensure /Compile/obj is used instead of <project>/obj
        # This is necessary because the files in obj/ will contain absolute paths, which are valid only in the container
        # Rider is okay with this, but Visual Studio throws a lot of errors if this happens
        directory_build_props = temp_files_directory / "Directory.Build.props"
        with directory_build_props.open("w+", encoding="utf-8") as file:
            file.write("""
<Project>
    <PropertyGroup>
        <BaseIntermediateOutputPath>/Compile/obj/$(MSBuildProjectName)/</BaseIntermediateOutputPath>
        <IntermediateOutputPath>/Compile/obj/$(MSBuildProjectName)/</IntermediateOutputPath>
        <DefaultItemExcludes>$(DefaultItemExcludes);backtests/*/code/**;live/*/code/**;optimizations/*/code/**</DefaultItemExcludes>
        <NoWarn>CS0618</NoWarn>
    </PropertyGroup>
</Project>
            """.strip())

        # Create a Python script that can be used to copy the right C# dependencies to /Lean/Launcher/bin/Debug
        # This script copies the correct DLLs even if a project has OS-specific DLLs
        copy_csharp_dependencies = temp_files_directory / "copy_csharp_dependencies.py"
        with copy_csharp_dependencies.open("w+", encoding="utf-8") as file:
            file.write("""
import json
import os
import platform
import shutil
import sys
from pathlib import Path

project_assets = json.loads(Path(sys.argv[-1]).read_text(encoding="utf-8"))
package_folders = [Path(folder) for folder in project_assets["packageFolders"].keys()]

ubuntu_version = os.popen("lsb_release -rs").read().strip()
accepted_runtime_identifiers = ["base", "unix", "linux", "debian", "ubuntu", f"ubuntu.{ubuntu_version}"]

if platform.machine() in ["arm64", "aarch64"]:
    accepted_runtime_identifiers.extend(["unix-arm", "linux-arm", "debian-arm", "ubuntu-arm", f"ubuntu.{ubuntu_version}-arm"])
else:
    accepted_runtime_identifiers.extend(["unix-x64", "linux-x64", "debian-x64", "ubuntu-x64", f"ubuntu.{ubuntu_version}-x64"])

def copy_file(library_id, partial_path, file_data):
    if "rid" in file_data and file_data["rid"] not in accepted_runtime_identifiers:
        return

    if not file_data.get("copyToOutput", True):
        return

    for folder in package_folders:
        full_path = folder / library_id.lower() / partial_path
        if full_path.exists():
            break
    else:
        return

    output_name = file_data.get("outputPath", full_path.name)

    target_path = Path("/Lean/Launcher/bin/Debug") / output_name
    if not target_path.exists():
        target_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(full_path, target_path)

project_target = list(project_assets["targets"].keys())[0]
for library_id, library_data in project_assets["targets"][project_target].items():
    for key, value in library_data.get("runtimeTargets", {}).items():
        copy_file(library_id, key, value)

    for key, value in library_data.get("runtime", {}).items():
        copy_file(library_id, key, value)

    for key, value in library_data.get("contentFiles", {}).items():
        copy_file(library_id, key, value)
            """.strip())

        run_options["mounts"].extend([Mount(target="/Directory.Build.props",
                                            source=str(directory_build_props),
                                            type="bind",
                                            read_only=False),
                                      Mount(target="/copy_csharp_dependencies.py",
                                            source=str(copy_csharp_dependencies),
                                            type="bind",
                                            read_only=False)])

    def _get_csharp_compile_root(self, project_dir: Path) -> Path:
        """Returns the path to the directory that should be mounted to compile the project directory.

        If the project is part of a solution this is the solution root.
        If the project is not part of a solution this is the project directory itself.

        :param project_dir: the path to the project directory
        :return: the path that should be mounted in the Docker container when compiling the C# project
        """
        current_dir = project_dir.parent
        while current_dir.parent != current_dir:
            if next(current_dir.glob("*.sln"), None) is not None:
                return current_dir

            current_dir = current_dir.parent

        return project_dir

    def _ensure_csproj_uses_correct_lean(self,
                                         compile_root: Path,
                                         csproj_path: Path,
                                         temp_dir: Path,
                                         run_options: Dict[str, Any]) -> None:
        """Ensures a C# project is compiled using the version of LEAN in the Docker container.

        When a .csproj file refers to the NuGet version of LEAN,
        we mount a temporary file on top of it which refers to the version of LEAN in the Docker container.
        In the case there is a breaking change between the two,
        this causes compiling to fail with readable error messages instead of ugly messages at runtime.

        :param compile_root: the path that is mounted in the Docker container
        :param csproj_path: the path to the .csproj file
        :param temp_dir: the temporary directory in which temporary .csproj files should be placed
        :param run_options: the dictionary to append run options to
        """
        csproj = self._xml_manager.parse(csproj_path.read_text(encoding="utf-8"))
        include_added = False

        for package_reference in csproj.iter("PackageReference"):
            if not package_reference.get("Include", "").lower().startswith("quantconnect."):
                continue

            if include_added:
                package_reference.getparent().remove(package_reference)

            package_reference.clear()

            package_reference.tag = "Reference"
            package_reference.set("Include", f"{LEAN_ROOT_PATH}/*.dll")
            package_reference.append(self._xml_manager.parse("<Private>False</Private>"))

            include_added = True

        if not include_added:
            return

        new_csproj_file = temp_dir / csproj_path.relative_to(compile_root)
        new_csproj_file.parent.mkdir(parents=True, exist_ok=True)
        with new_csproj_file.open("w+", encoding="utf-8") as file:
            file.write(self._xml_manager.to_string(csproj))

        run_options["mounts"].append(Mount(target=f"/LeanCLI/{csproj_path.relative_to(compile_root).as_posix()}",
                                           source=str(new_csproj_file),
                                           type="bind",
                                           read_only=True))

    def _force_disk_provider_if_necessary(self,
                                          lean_config: Dict[str, Any],
                                          config_key: str,
                                          zip_provider: str,
                                          disk_provider: str,
                                          zip_dir: Path) -> None:
        """Updates the Lean config to use the disk provider instead of the zip one if there are no zips to use.

        :param lean_config: the Lean config to update
        :param config_key: the key of the configuration property
        :param zip_provider: the fully classified name of the zip provider for this property
        :param disk_provider: the fully classified name of the disk provider for this property
        :param zip_dir: the directory where the zip provider looks for zip files
        """
        if lean_config.get(config_key, None) != zip_provider:
            return

        if not zip_dir.exists():
            lean_config[config_key] = disk_provider
            return

        zip_names = sorted([f.name for f in zip_dir.iterdir() if f.name.endswith(".zip")], reverse=True)
        zip_names = [re.sub(r"[^\d]", "", name) for name in zip_names]

        if len(zip_names) == 0 or (datetime.now() - datetime.strptime(zip_names[0], "%Y%m%d")).days > 7:
            lean_config[config_key] = disk_provider

    def setup_language_specific_run_options(self, run_options, project_dir, algorithm_file,
                                            set_up_common_csharp_options_called, release) -> None:
        # Set up language-specific run options
        if algorithm_file.name.endswith(".py"):
            self.set_up_python_options(project_dir, run_options)
        else:
            if not set_up_common_csharp_options_called:
                self.set_up_common_csharp_options(run_options)
            self.set_up_csharp_options(project_dir, run_options, release)

    def format_error_before_logging(self, chunk: str):
        from lean.components.util import compiler
        errors = []

        # As we don't have algorithm type information. We can check errors for both
        # Python
        jsonString = compiler.get_errors("python", chunk)
        jsonData = json.loads(jsonString)
        errors.extend(jsonData["aErrors"])
        # CSharp
        jsonString = compiler.get_errors("csharp", chunk)
        jsonData = json.loads(jsonString)
        errors.extend(jsonData["aErrors"])

        for error in errors:
            self._logger.info(error)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Any
from lean.models.json_module import JsonModule
from lean.models.configuration import InternalInputUserInput, TradingEnvConfiguration


class CloudBrokerage(JsonModule):
    """A JsonModule implementation for the cloud brokerages."""

    def __init__(self, json_cloud_brokerage_data: Dict[str, Any]) -> None:
        super().__init__(json_cloud_brokerage_data)

    def get_id(self) -> str:
        """Returns the id of the brokerage.
        :return: the id of this brokerage as it is expected by the live/create API endpoint
        """
        return self._id

    def _get_settings(self) -> Dict[str, str]:
        """Returns all settings for this brokerage, except for the id.
        :return: the settings of this brokerage excluding the id
        """
        settings = {}
        for config in self.get_required_configs():
            value = None
            if not config._cloud_id:
                continue
            # TODO: handle cases where tranding env config is not present, environment will still be required.
            if type(config) == TradingEnvConfiguration:
                value = "paper" if str(config).lower() in [
                    "practice", "demo", "beta", "paper"] else "live"
            elif type(config) is InternalInputUserInput:
                if not config._is_conditional:
                    value = config._value
                else:
                    for option in config._value_options:
                        if option._condition.check(self.get_config_value_from_name(option._condition._dependent_config_id)):
                            value = option._value
                            break
                    if not value:
                        raise ValueError(
                            f'No condtion matched among present options for {config._cloud_id}')
            else:
                value = config._value
            settings[config._cloud_id] = value
        return settings

    def get_settings(self) -> Dict[str, str]:
        """Returns all settings for this brokerage.
        :return: the settings to set in the "brokerage" property of the live/create API endpoint
        """
        settings = self._get_settings()
        if "environment" not in settings.keys():
            settings["environment"] = "live"
        settings["id"] = self.get_id()
        return settings

    def get_price_data_handler(self) -> str:
        """Returns the price data feed handler to use.
        :return: the value to assign to the "dataHandler" property of the live/create API endpoint
        """
        # TODO: Handle this case with json conditions
        if self.get_name() == "Interactive Brokers":
            return "InteractiveBrokersHandler" if self.get_config_value_from_name("ib-data-feed") else "QuantConnectHandler"
        return "QuantConnectHandler"



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shutil
import subprocess
from pathlib import Path
from typing import Union

import click
from pkg_resources import Requirement

from lean.click import LeanCommand, PathParameter
from lean.container import container
from lean.models.errors import MoreInfoError


def _remove_package_from_csharp_project(project_dir: Path, name: str, no_local: bool) -> None:
    """Removes a custom C# library from a C# project.

    Removes the library from the project's .csproj file,
    and restores the project if dotnet is on the user's PATH and no_local is False.

    :param project_dir: the path to the project directory
    :param name: the name of the library to remove
    :param no_local:
    """
    logger = container.logger()
    path_manager = container.path_manager()

    csproj_file = next(p for p in project_dir.iterdir() if p.name.endswith(".csproj"))
    logger.info(f"Removing {name} from '{path_manager.get_relative_path(csproj_file)}'")

    xml_manager = container.xml_manager()
    csproj_tree = xml_manager.parse(csproj_file.read_text(encoding="utf-8"))

    for package_reference in csproj_tree.findall(".//PackageReference"):
        if package_reference.get("Include", "").lower() == name.lower():
            package_reference.getparent().remove(package_reference)

    csproj_file.write_text(xml_manager.to_string(csproj_tree), encoding="utf-8")

    if not no_local and shutil.which("dotnet") is not None:
        logger.info(f"Restoring packages in '{path_manager.get_relative_path(project_dir)}'")

        process = subprocess.run(["dotnet", "restore", str(csproj_file)], cwd=project_dir)

        if process.returncode != 0:
            raise RuntimeError("Something went wrong while restoring packages, see the logs above for more information")


def _remove_pypi_package_from_python_project(project_dir: Path, name: str) -> None:
    """Removes a custom Python library from a Python project.

    Removes the library from the project's requirements.txt file.

    :param project_dir: the path to the project directory
    :param name: the name of the library to remove
    """
    logger = container.logger()
    path_manager = container.path_manager()

    requirements_file = project_dir / "requirements.txt"
    logger.info(f"Removing {name} from '{path_manager.get_relative_path(requirements_file)}'")

    if not requirements_file.is_file():
        return

    requirements_content = requirements_file.read_text(encoding="utf-8")
    new_lines = []

    for line in requirements_content.splitlines():
        try:
            requirement = Requirement.parse(line)
            if requirement.name.lower() != name.lower():
                new_lines.append(line)
        except ValueError:
            new_lines.append(line)

    new_content = "\n".join(new_lines).strip()
    new_content = new_content + "\n" if len(new_content) > 0 else new_content
    requirements_file.write_text(new_content, encoding="utf-8")


@click.command(cls=LeanCommand, requires_docker=True)
@click.argument("project", type=PathParameter(exists=True, file_okay=False, dir_okay=True))
@click.argument("name", type=str)
@click.option("--no-local", is_flag=True, default=False, help="Skip making changes to your local environment")
def remove(project: Path, name: str, no_local: bool) -> None:
    """Remove a custom library from a project.

    PROJECT must be the path to the project directory.

    NAME must be either the name of the NuGet package (for C# projects), the PyPI package (for Python projects),
    or the path to the Lean CLI library to remove.

    Custom C# libraries are removed from the project's .csproj file,
    which is then restored if dotnet is on your PATH and the --no-local flag has not been given.

    Custom Python libraries are removed from the project's requirements.txt file.

    \b
    C# example usage:
    $ lean library remove "My CSharp Project" Microsoft.ML

    \b
    Python example usage:
    $ lean library remove "My Python Project" tensorflow
    """
    project_config = container.project_config_manager().get_project_config(project)
    project_language = project_config.get("algorithm-language", None)

    if project_language is None:
        raise MoreInfoError(f"{project} is not a Lean CLI project",
                            "https://www.lean.io/docs/v2/lean-cli/projects/project-management#02-Create-Projects")

    library_manager = container.library_manager()
    library_dir = Path(name).expanduser().resolve()
    if library_manager.is_lean_library(library_dir):
        if project_language == "CSharp":
            library_manager.remove_lean_library_from_csharp_project(project, library_dir, no_local)
        else:
            library_manager.remove_lean_library_from_python_project(project, library_dir)
    else:
        if project_language == "CSharp":
            _remove_package_from_csharp_project(project, name, no_local)
        else:
            _remove_pypi_package_from_python_project(project, name)
